"""
Author: Shadi Zabad
Date: December 2020
"""

from typing import Union, Dict

import warnings
import copy
import pandas as pd
import numpy as np
from tqdm import tqdm

from .GenotypeMatrix import GenotypeMatrix, xarrayGenotypeMatrix, plinkBEDGenotypeMatrix
from .SampleTable import SampleTable
from .SumstatsTable import SumstatsTable
from .AnnotationMatrix import AnnotationMatrix
from .LDMatrix import LDMatrix

from .utils.compute_utils import iterable
from .utils.system_utils import makedir, get_filenames


class GWADataLoader(object):

    def __init__(self,
                 bed_files=None,
                 phenotype_file=None,
                 covariates_file=None,
                 keep_samples=None,
                 keep_file=None,
                 extract_snps=None,
                 extract_file=None,
                 min_maf=None,
                 min_mac=None,
                 drop_duplicated=True,
                 phenotype_likelihood='gaussian',
                 sumstats_files=None,
                 sumstats_format='magenpy',
                 ld_store_files=None,
                 annotation_files=None,
                 backend='xarray',
                 temp_dir='temp',
                 output_dir='output',
                 verbose=True,
                 n_threads=1):

        # ------- Sanity checks -------

        assert backend in ('xarray', 'plink')
        assert phenotype_likelihood in ('gaussian', 'binomial')

        # ------- General options -------

        self.backend = backend

        self.temp_dir = temp_dir
        self.output_dir = output_dir
        self.cleanup_dir_list = []  # Directories to clean up after execution.

        makedir([temp_dir, output_dir])

        self.verbose = verbose
        self.n_threads = n_threads

        # ------- General parameters -------

        self.phenotype_likelihood: str = phenotype_likelihood

        self.genotype: Union[Dict[int, GenotypeMatrix], None] = None
        self.sample_table: Union[SampleTable, None] = None
        self.ld: Union[Dict[int, LDMatrix], None] = None
        self.sumstats_table: Union[Dict[int, SumstatsTable], None] = None
        self.annotation: Union[Dict[int, AnnotationMatrix], None] = None

        # ------- Read data files -------

        self.read_genotypes(bed_files,
                            min_maf=min_maf,
                            min_mac=min_mac,
                            drop_duplicated=drop_duplicated)
        self.read_phenotype(phenotype_file)
        self.read_covariates(covariates_file)
        self.read_ld(ld_store_files)
        self.read_annotations(annotation_files)
        self.read_summary_statistics(sumstats_files,
                                     sumstats_format,
                                     drop_duplicated=drop_duplicated)

        # ------- Filter samples or SNPs -------

        if extract_snps is not None or extract_file is not None:
            self.filter_snps(extract_snps=extract_snps, extract_file=extract_file)

        if keep_samples is not None or keep_file is not None:
            self.filter_samples(keep_samples=keep_samples, keep_file=keep_file)

        # ------- Harmonize data sources -------

        self.harmonize_data()

    @property
    def samples(self):
        if self.sample_table is not None:
            return self.sample_table.iid

    @property
    def sample_size(self):
        """
        The number of samples.
        """
        if self.sample_table is not None:
            return self.sample_table.n
        elif self.sumstats_table is not None:
            return np.max([np.max(ss.n_per_snp) for ss in self.sumstats_table.values()])
        else:
            raise Exception("Information about the sample size is not available!")

    @property
    def n(self):
        """
        The number of samples. See also `.sample_size()`.
        """

        return self.sample_size

    @property
    def snps(self):
        """
        Return the list of SNPs retained in each chromosome.
        """
        if self.genotype is not None:
            return {c: g.snps for c, g in self.genotype.items()}
        elif self.sumstats_table is not None:
            return {c: s.snps for c, s in self.sumstats_table.items()}
        elif self.ld is not None:
            return {c: l.snps for c, l in self.ld.items()}
        elif self.annotation is not None:
            return {c: a.snps for c, a in self.annotation.items()}
        else:
            raise Exception("GWADataLoader is not properly initialized!")

    @property
    def m(self):
        """
        The number of variants. See also `.n_snps`
        """
        return sum(self.shapes.values())

    @property
    def n_snps(self):
        """
        The number of variants. See also `.m`
        """
        return self.m

    @property
    def shapes(self):
        """
        Return a dictionary where the key is the chromosome number and the value is
        the number of variants on that chromosome.
        """
        if self.genotype is not None:
            return {c: g.shape[1] for c, g in self.genotype.items()}
        elif self.sumstats_table is not None:
            return {c: s.shape[0] for c, s in self.sumstats_table.items()}
        elif self.ld is not None:
            return {c: l.shape[0] for c, l in self.ld.items()}
        elif self.annotation is not None:
            return {c: a.shape[0] for c, a in self.annotation.items()}
        else:
            raise Exception("GWADataLoader is not properly initialized!")

    @property
    def chromosomes(self):
        """
        Return the list of chromosomes that were loaded to `GWADataLoader`.
        """
        return sorted(list(self.shapes.keys()))

    @property
    def n_annotations(self):
        """
        Return the number of annotations included in the annotation matrices.
        """
        if self.annotation is not None:
            return self.annotation[self.chromosomes[0]].n_annotations

    def filter_snps(self, extract_snps=None, extract_file=None, chromosome=None):
        """
        Filter the SNP set from all the GWADataLoader objects.
        :param extract_snps: A list or vector of SNP IDs to keep.
        :param extract_file: A path to a plink-style file with SNP IDs to keep.
        :param chromosome: Chromosome number. If specified, applies the filter on that chromosome only.
        """

        if extract_snps is None and extract_file is None:
            return

        if chromosome is not None:
            chroms = [chromosome]
        else:
            chroms = self.chromosomes

        if extract_snps is None:
            from .parsers.misc_parsers import read_snp_filter_file
            extract_snps = read_snp_filter_file(extract_file)

        for c in chroms:

            # Filter the genotype matrix:
            if self.genotype is not None and c in self.genotype:
                self.genotype[c].filter_snps(extract_snps=extract_snps)

            # Filter the summary statistics table:
            if self.sumstats_table is not None and c in self.sumstats_table:
                self.sumstats_table[c].filter_snps(extract_snps=extract_snps)

            if self.ld is not None and c in self.ld:
                self.ld[c].filter_snps(extract_snps=extract_snps)

            # Filter the annotation matrix:
            if self.annotation is not None and c in self.annotation:
                self.annotation[c].filter_snps(extract_snps=extract_snps)

    def filter_samples(self, keep_samples=None, keep_file=None):
        """
        Filter samples from the samples table. User must specify
        either a list of samples to keep or the path to a file
        with the list of samples to keep.

        :param keep_samples: A list (or array) of sample IDs to keep.
        :param keep_file: The path to a file with the list of samples to keep.
        """

        self.sample_table.filter_samples(keep_samples=keep_samples, keep_file=keep_file)
        self.sync_sample_tables()

    def read_annotations(self, annot_paths):
        """
        Read the annotation matrix from file. Annotations are a set of features associated
        with each SNP and are generally represented in table format. Here, we support tables
        of the form developed and standardized by the LDSC software.
        Consult the documentation for `AnnotationMatrix` for more details.

        :param annot_paths: The path to the annotation file(s). You may use a wildcard here to
        read files for multiple chromosomes.
        """

        if annot_paths is None:
            return

        # Find all the relevant files in the path passed by the user:
        if not iterable(annot_paths):
            annot_files = get_filenames(annot_paths, extension='.annot')
        else:
            annot_files = annot_paths

        if len(annot_files) < 1:
            warnings.warn(f"No annotation files were found at: {annot_paths}")
            return

        if self.verbose and len(annot_files) < 2:
            print("> Reading annotation file...")

        self.annotation = {}

        for annot_file in tqdm(annot_files,
                               total=len(annot_files),
                               desc="Reading annotation files",
                               disable=not self.verbose or len(annot_files) < 2):
            annot_mat = AnnotationMatrix.from_file(annot_file)
            self.annotation[annot_mat.chromosome] = annot_mat

    def read_genotypes(self,
                       bed_paths,
                       keep_samples=None,
                       keep_file=None,
                       extract_snps=None,
                       extract_file=None,
                       min_maf=None,
                       min_mac=1,
                       drop_duplicated=True):
        """
        Read the genotype matrix and/or associated metadata from plink's BED file format.
        Consult the documentation for `GenotypeMatrix` for more details.

        :param bed_paths: The path to the BED file(s). You may use a wildcard here to read files for multiple
        chromosomes.
        :param keep_samples: A vector or list of sample IDs to keep when filtering the genotype matrix.
        :param keep_file: A path to a plink-style file containing sample IDs to keep.
        :param extract_snps: A vector or list of SNP IDs to keep when filtering the genotype matrix.
        :param extract_file: A path to a plink-style file containing SNP IDs to keep.
        :param min_maf: The minimum minor allele frequency cutoff.
        :param min_mac: The minimum minor allele count cutoff.
        :param drop_duplicated: If True, drop SNPs with duplicated rsID.
        """

        if bed_paths is None:
            return

        # Find all the relevant files in the path passed by the user:
        if not iterable(bed_paths):
            bed_files = get_filenames(bed_paths, extension='.bed')
        else:
            bed_files = bed_paths

        if len(bed_files) < 1:
            warnings.warn(f"No BED files were found at: {bed_paths}")
            return

        # Depending on the backend, select the `GenotypeMatrix` class:
        if self.backend == 'xarray':
            gmat_class = xarrayGenotypeMatrix
        else:
            gmat_class = plinkBEDGenotypeMatrix

        if self.verbose and len(bed_files) < 2:
            print("> Reading BED file...")

        self.genotype = {}

        for bfile in tqdm(bed_files,
                          total=len(bed_files),
                          desc="Reading BED files",
                          disable=not self.verbose or len(bed_files) < 2):
            # Read BED file and update the genotypes dictionary:
            self.genotype.update(gmat_class.from_file(bfile, temp_dir=self.temp_dir).split_by_chromosome())

        # After reading the genotype matrices, apply some standard filters:
        for i, (c, g) in enumerate(self.genotype.items()):

            # Filter the genotype matrix to keep certain subsample:
            if keep_samples or keep_file:
                g.filter_samples(keep_samples=keep_samples, keep_file=keep_file)

            # Filter the genotype matrix to keep certain SNPs
            if extract_snps or extract_file:
                g.filter_snps(extract_snps=extract_snps, extract_file=extract_file)

            # Drop duplicated SNP IDs
            if drop_duplicated:
                g.drop_duplicated_snps()

            # Filter SNPs by minor allele frequency and/or count:
            g.filter_by_allele_frequency(min_maf=min_maf, min_mac=min_mac)

            if i == 0:
                self.sample_table = g.sample_table

    def read_phenotype(self, phenotype_file, drop_na=True, **read_csv_kwargs):
        """
        Read the phenotype file and integrate it with the sample tables and genotype matrices.

        :param phenotype_file: The path to the phenotype file
        (Default: tab-separated file with `FID IID phenotype` columns).
        :param drop_na: Drop samples with missing phenotype information.
        :param read_csv_kwargs: keyword arguments for the `read_csv` function of `pandas`.
        """

        if phenotype_file is None:
            return

        if self.verbose:
            print("> Reading phenotype file...")

        assert self.sample_table is not None

        self.sample_table.read_phenotype_file(phenotype_file, drop_na=drop_na, **read_csv_kwargs)
        self.sync_sample_tables()

    def set_phenotype(self, new_phenotype, phenotype_likelihood=None):
        """
        A convenience method to update the phenotype column for the samples.
        :param new_phenotype: A vector or list of phenotype values.
        :param phenotype_likelihood: The phenotype likelihood (e.g. `binomial`, `gaussian`). Optional.
        """

        self.sample_table.set_phenotype(new_phenotype,
                                        phenotype_likelihood=phenotype_likelihood or self.phenotype_likelihood)
        self.sync_sample_tables()

    def read_covariates(self, covariates_file, **read_csv_kwargs):
        """
        Read the covariates file and integrate it with the sample tables and genotype matrices.

        :param covariates_file: The path to the covariates file
        (Default: tab-separated file starting with the `FID IID ...` columns and followed by the covariate columns).
        :param read_csv_kwargs: keyword arguments for the `read_csv` function of `pandas`.
        """

        if covariates_file is None:
            return

        if self.verbose:
            print("> Reading covariates file...")

        assert self.sample_table is not None

        self.sample_table.read_covariates_file(covariates_file, **read_csv_kwargs)
        self.sync_sample_tables()

    def read_summary_statistics(self,
                                sumstats_path,
                                sumstats_format='magenpy',
                                parser=None,
                                drop_duplicated=True,
                                **parse_kwargs):
        """
        Read GWAS summary statistics file(s) and parse them to `SumstatsTable` objects.

        :param sumstats_path: The path to the summary statistics file(s). The path may be a wildcard.
        :param sumstats_format: The format for the summary statistics. Currently supports the following
         formats: `plink`, `magenpy`, `fastGWA`, `COJO`.
        :param parser: If the summary statistics file does not follow any of the formats above, you can create
        your own parser by inheriting from the base `SumstatsParser` class and passing it here as an argument.
        :param drop_duplicated: Drop SNPs with duplicated rsIDs.
        :param parse_kwargs: keyword arguments for the parser. These are mainly parameters that will be passed to
        `pandas.read_csv` function, such as the delimiter, header information, etc.
        """

        if sumstats_path is None:
            return

        if not iterable(sumstats_path):
            sumstats_files = get_filenames(sumstats_path)

            from .utils.system_utils import valid_url
            if len(sumstats_files) < 1 and valid_url(sumstats_path):
                sumstats_files = [sumstats_path]
        else:
            sumstats_files = sumstats_path

        if len(sumstats_files) < 1:
            warnings.warn(f"No summary statistics files were found at: {sumstats_path}")
            return

        if self.verbose and len(sumstats_files) < 2:
            print("> Reading summary statistics file...")

        self.sumstats_table = {}

        for f in tqdm(sumstats_files,
                      total=len(sumstats_files),
                      desc="Reading summary statistics files",
                      disable=not self.verbose or len(sumstats_files) < 2):

            ss_tab = SumstatsTable.from_file(f, sumstats_format=sumstats_format, parser=parser, **parse_kwargs)

            if drop_duplicated:
                ss_tab.drop_duplicates()

            if 'CHR' in ss_tab.table.columns:
                self.sumstats_table.update(ss_tab.split_by_chromosome())
            else:
                if self.genotype is not None:
                    ref_table = {c: g.snps for c, g in self.genotype.items()}
                elif self.ld is not None:
                    ref_table = {c: ld.snps for c, ld in self.ld.items()}
                else:
                    raise Exception("Cannot index summary statistics tables without chromosome information!")

                self.sumstats_table.update(ss_tab.split_by_chromosome(snps_per_chrom=ref_table))

    def read_ld(self, ld_store_paths):
        """
        Read the LD matrix files stored on-disk in Zarr array format.
        :param ld_store_paths: The path to the LD matrices. This may be a wildcard to accommodate reading data
        for multiple chromosomes.
        """

        if ld_store_paths is None:
            return

        if not iterable(ld_store_paths):
            ld_store_files = get_filenames(ld_store_paths, extension='.zarray')
        else:
            ld_store_files = ld_store_paths

        if len(ld_store_files) < 1:
            warnings.warn(f"No LD matrix files were found at: {ld_store_paths}")
            return

        if self.verbose and len(ld_store_files) < 2:
            print("> Reading LD matrix...")

        self.ld = {}

        for f in tqdm(ld_store_files,
                      total=len(ld_store_files),
                      desc="Reading LD matrices",
                      disable=not self.verbose or len(ld_store_files) < 2):
            z = LDMatrix.from_path(f)
            self.ld[z.chromosome] = z

    def load_ld(self):
        """
        A utility function to load the LD matrices to memory from on-disk storage.
        """
        if self.ld is not None:
            for ld in self.ld.values():
                ld.load()

    def release_ld(self):
        """
        A utility function to release LD matrices from memory.
        """
        if self.ld is not None:
            for ld in self.ld.values():
                ld.release()

    def compute_ld(self, estimator, output_dir, **ld_kwargs):
        """
        Compute the Linkage-Disequilibrium (LD) matrix or SNP-by-SNP Pearson
        correlation matrix between genetic variants. This function only considers correlations
        between SNPs on the same chromosome. This is a utility function that calls the
        `.compute_ld()` method of the `GenotypeMatrix` objects associated with
        GWADataLoader.

        :param estimator: The estimator for the LD matrix. We currently support
        4 different estimators: `sample`, `windowed`, `shrinkage`, and `block`.
        :param output_dir: The output directory where the Zarr array containing the
        entries of the LD matrix will be stored.
        :param ld_kwargs: keyword arguments for the various LD estimators. Consult
        the implementations of `WindowedLD`, `ShrinkageLD`, and `BlockLD` for details.
        """

        if self.verbose and len(self.genotype) < 2:
            print("> Computing LD matrix...")

        self.ld = {
            c: g.compute_ld(estimator, output_dir, **ld_kwargs)
            for c, g in tqdm(sorted(self.genotype.items(), key=lambda x: x[0]),
                             total=len(self.genotype),
                             desc='Computing LD matrices',
                             disable=not self.verbose or len(self.genotype) < 2)
        }

    def get_ld_matrices(self):
        return self.ld

    def get_ld_boundaries(self):

        if self.ld is None:
            return None

        return {c: ld.get_masked_boundaries() for c, ld in self.ld.items()}

    def harmonize_data(self):
        """
        This method ensures that the data sources (reference genotype,
        LD matrices, summary statistics, annotations) are all aligned in terms of the
        set of variants that they operate on as well as the designation of the effect allele for
        each variant.
        """

        data_sources = (self.genotype, self.sumstats_table, self.ld, self.annotation)
        initialized_data_sources = [ds for ds in data_sources if ds is not None]

        # If less than two data sources are present, skip harmonization...
        if len(initialized_data_sources) < 2:
            return

        # Get the chromosomes information from all the data sources:
        chromosomes = list(set.union(*[set(ds.keys()) for ds in initialized_data_sources]))

        if self.verbose and len(chromosomes) < 2:
            print("> Harmonizing data...")

        for c in tqdm(chromosomes,
                      total=len(chromosomes),
                      desc='Harmonizing data',
                      disable=not self.verbose or len(chromosomes) < 2):

            # Which initialized data sources have information for chromosome `c`
            miss_chroms = [c not in ds for ds in initialized_data_sources]

            if sum(miss_chroms) > 0:
                # If the chromosome data only exists for some data sources but not others, remove the chromosome
                # from all data source.
                # Is this the best way to handle the missingness? Should we just post a warning?
                for ds in initialized_data_sources:
                    if c in ds:
                        del ds[c]

            else:

                # Find the set of SNPs that are shared across all data sources:
                common_snps = np.array(list(set.intersection(*[set(ds[c].snps)
                                                               for ds in initialized_data_sources])))

                # If necessary, filter the data sources to only have the common SNPs:
                for ds in initialized_data_sources:
                    if ds[c].n_snps != len(common_snps):
                        ds[c].filter_snps(extract_snps=common_snps)

                # Harmonize the summary statistics data with either genotype or LD reference.
                # This procedure checks for flips in the effect allele between data sources.
                if self.sumstats_table is not None:
                    if self.genotype is not None:
                        self.sumstats_table[c].match(self.genotype[c].get_snp_table(col_subset=['SNP', 'A1', 'A2']))
                    elif self.ld is not None:
                        self.sumstats_table[c].match(self.ld[c].to_snp_table(col_subset=['SNP', 'A1', 'A2']))

                    # If during the allele matching process we discover incompatibilities,
                    # we filter those SNPs:
                    for ds in initialized_data_sources:
                        if ds[c].n_snps != self.sumstats_table[c].n_snps:
                            ds[c].filter_snps(extract_snps=self.sumstats_table[c].snps)

    def perform_gwas(self, **gwa_kwargs):
        """
        Perform genome-wide association testing of all variants against the phenotype.
        This is a utility function that calls the `.perform_gwas()` method of the
        `GenotypeMatrix` objects associated with GWADataLoader.

        :param gwa_kwargs: Keyword arguments to pass to the GWA functions. Consult stats.gwa.utils
        for relevant keyword arguments for each backend.
        """

        if self.verbose and len(self.genotype) < 2:
            print("> Performing GWAS...")

        self.sumstats_table = {
            c: g.perform_gwas(**gwa_kwargs)
            for c, g in tqdm(sorted(self.genotype.items(), key=lambda x: x[0]),
                             total=len(self.genotype),
                             desc='Performing GWAS',
                             disable=not self.verbose or len(self.genotype) < 2)
        }

    def score(self, beta=None):
        """
        Perform linear scoring, i.e. multiply the genotype matrix by the vector of effect sizes, `beta`.

        :param beta: A dictionary where the keys are the chromosome numbers and the
        values are a vector of effect sizes for each variant on that chromosome. If the
        betas are not provided, we use the marginal betas by default (if those are available).
        """

        if beta is None:
            try:
                beta = {c: s.marginal_beta or s.get_snp_pseudo_corr() for c, s in self.sumstats_table.items()}
            except Exception:
                raise Exception("To perform linear scoring, you must a provide effect size estimates (BETA)!")

        common_chroms = sorted(list(set(self.genotype.keys()).intersection(set(beta.keys()))))

        if self.verbose and len(common_chroms) < 2:
            print("> Generating polygenic scores...")

        pgs = None

        for c in tqdm(common_chroms,
                      total=len(common_chroms),
                      desc='Generating polygenic scores',
                      disable=not self.verbose or len(common_chroms) < 2):

            if pgs is None:
                pgs = self.genotype[c].score(beta[c])
            else:
                pgs += self.genotype[c].score(beta[c])

        # If we only have a single set of betas, flatten the PGS vector:
        if len(pgs.shape) > 1:
            if pgs.shape[1] == 1:
                pgs = pgs.flatten()

        return pgs

    def predict(self, beta=None):
        """
        Predict the phenotype for the genotyped samples using the provided effect size
        estimates `beta`. For quantitative traits, this is equivalent to performing
        linear scoring. For binary phenotypes, we transform the output using probit link function.

        :param beta: A dictionary where the keys are the chromosome numbers and the
        values are a vector of effect sizes for each variant on that chromosome. If the
        betas are not provided, we use the marginal betas by default (if those are available).
        """

        # Perform linear scoring:
        pgs = self.score(beta)

        if self.phenotype_likelihood == 'binomial':
            # Apply probit link function:
            from scipy.stats import norm
            pgs = norm.cdf(pgs)

        return pgs

    def to_individual_table(self):
        """
        Get a plink-style dataframe of individual IDs, in the form of
        Family ID (FID) and Individual ID (IID).
        """

        return self.sample_table.get_individual_table()

    def to_phenotype_table(self):
        """
        Get a plink-style dataframe with each individual's Family ID (FID),
        Individual ID (IID), and phenotype value.
        """

        return self.sample_table.get_phenotype_table()

    def to_snp_table(self, col_subset=None, per_chromosome=False):
        """
        Return a dataframe of SNP information for all variants
        across different chromosomes.

        :param col_subset: The subset of columns to obtain.
        :param per_chromosome: If True, returns a dictionary where the key
        is the chromosome number and the value is the SNP table per
        chromosome.
        """

        snp_tables = {}

        for c in self.chromosomes:
            if self.sumstats_table is not None:
                snp_tables[c] = self.sumstats_table[c].get_table(col_subset=col_subset)
            elif self.genotype is not None:
                snp_tables[c] = self.genotype[c].get_snp_table(col_subset=col_subset)
            elif self.ld is not None:
                snp_tables[c] = self.ld[c].to_snp_table(col_subset=col_subset)
            else:
                raise Exception("GWADataLoader is not properly initialized!")

        if per_chromosome:
            return snp_tables
        else:
            return pd.concat(list(snp_tables.values()))

    def to_summary_statistics_table(self, col_subset=None, per_chromosome=False):
        """
        Return a dataframe of the GWAS summary statistics for all variants
        across different chromosomes.

        :param col_subset: The subset of columns (or summary statistics) to obtain.
        :param per_chromosome: If True, returns a dictionary where the key
        is the chromosome number and the value is the summary statistics table per
        chromosome.
        """

        assert self.sumstats_table is not None

        snp_tables = {}

        for c in self.chromosomes:
            snp_tables[c] = self.sumstats_table[c].get_table(col_subset=col_subset)

        if per_chromosome:
            return snp_tables
        else:
            return pd.concat(list(snp_tables.values()))

    def sync_sample_tables(self):
        """
        A utility method to sync the sample tables of the
        `GenotypeMatrix` objects with the sample table under
        the `GWADataLoader` object. This is especially important
        when setting new phenotypes (from the simulators) or reading
        covariates files, etc.
        """

        for c, g in self.genotype.items():
            g.set_sample_table(self.sample_table)

    def split_by_chromosome(self):
        """
        A utility method to split a GWADataLoader object by chromosome ID, such that
        we would have one `GWADataLoader` object per chromosome. The method returns a dictionary
        where the key is the chromosome number and the value is the `GWADataLoader` object corresponding
        to that chromosome only.
        """

        if len(self.chromosomes) == 1:
            return {self.chromosomes[0]: self}

        else:
            split_dict = {}

            for c in self.chromosomes:
                split_dict[c] = copy.copy(self)

                if self.genotype is not None and c in self.genotype:
                    split_dict[c].genotype = {c: self.genotype[c]}
                if self.sumstats_table is not None and c in self.sumstats_table:
                    split_dict[c].sumstats_table = {c: self.sumstats_table[c]}
                if self.ld is not None and c in self.ld:
                    split_dict[c].ld = {c: self.ld[c]}
                if self.annotation is not None and c in self.annotation:
                    split_dict[c].annotation = {c: self.annotation[c]}

            return split_dict

    def split_by_samples(self, proportions=None, groups=None, keep_original=False):
        """
        Split the `GWADataLoader` object by samples, if genotype or sample data
        is available. The user must provide a list or proportion of samples in each split,
        and the method will return a list of `GWADataLoader` objects with only samples
        within each split. This may be a useful utility for training/testing split or some
        other downstream tasks.

        :param proportions: A list with the proportion of samples in each split. Must add to 1.
        :param groups: A list of lists containing the sample IDs in each split.
        :param keep_original: If True, keep the original `GWADataLoader` object and do not
        transform it in the splitting process.
        """

        if self.sample_table is None:
            raise Exception("The sample table is not set!")

        if groups is None:
            if proportions is None:
                raise Exception("To split a `GWADataloader` object by samples, the user must provide either the list "
                                "or proportion of individuals in each split.")
            else:

                # Assign each sample to a different split randomly by drawing from a multinomial:
                random_split = np.random.multinomial(1, proportions, size=self.sample_size).astype(bool)
                # Extract the samples in each group from the multinomial sample:
                groups = [self.samples[random_split[:, i]] for i in range(random_split.shape[1])]

        gdls = []
        for i, g in enumerate(groups):

            if len(g) < 1:
                raise Exception(f"Group {i} is empty! Please ensure that all splits have at least one sample.")

            if (i + 1) == len(groups) and not keep_original:
                new_gdl = self
            else:
                new_gdl = copy.deepcopy(self)

            new_gdl.filter_samples(keep_samples=g)

            gdls.append(new_gdl)

        return gdls

    def cleanup(self):
        """
        Clean up all temporary files and directories
        """
        if self.verbose:
            print("> Cleaning up workspace.")

        for tmpdir in self.cleanup_dir_list:
            try:
                tmpdir.cleanup()
            except FileNotFoundError:
                continue

        # Clean up the temporary files associated with the genotype matrices:
        if self.genotype is not None:
            for g in self.genotype.values():
                g.cleanup()

        # Release the LD data from memory:
        self.release_ld()



import os.path as osp
import pandas as pd
from magenpy.utils.executors import plink2Executor
from magenpy.GenotypeMatrix import plinkBEDGenotypeMatrix
from magenpy.utils.model_utils import merge_snp_tables


def compute_allele_frequency_plink2(genotype_matrix, temp_dir='temp'):

    assert isinstance(genotype_matrix, plinkBEDGenotypeMatrix)

    plink2 = plink2Executor()

    s_table = genotype_matrix.sample_table

    keep_file = osp.join(temp_dir, 'samples.keep')
    keep_table = s_table.get_individual_table()
    keep_table.to_csv(keep_file, index=False, header=False, sep="\t")

    snp_keepfile = osp.join(temp_dir, f"variants.keep")
    pd.DataFrame({'SNP': genotype_matrix.snps}).to_csv(
        snp_keepfile, index=False, header=False
    )

    plink_output = osp.join(temp_dir, "variants")

    cmd = [
        f"--bfile {genotype_matrix.bed_file}",
        f"--keep {keep_file}",
        f"--extract {snp_keepfile}",
        f"--freq",
        f"--out {plink_output}",
    ]

    plink2.execute(cmd)

    freq_df = pd.read_csv(plink_output + ".afreq", delim_whitespace=True)
    freq_df.rename(columns={'ID': 'SNP',
                            'REF': 'A2',
                            'ALT': 'A1', 'ALT1': 'A1',
                            'ALT_FREQS': 'MAF', 'ALT1_FREQ': 'MAF'}, inplace=True)
    merged_df = merge_snp_tables(genotype_matrix.get_snp_table(['SNP', 'A1', 'A2']), freq_df)

    if len(merged_df) != genotype_matrix.n_snps:
        raise ValueError("Length of allele frequency table does not match number of SNPs.")

    return merged_df['MAF'].values


def compute_sample_size_per_snp_plink2(genotype_matrix, temp_dir='temp'):

    assert isinstance(genotype_matrix, plinkBEDGenotypeMatrix)

    plink2 = plink2Executor()

    s_table = genotype_matrix.sample_table

    keep_file = osp.join(temp_dir, 'samples.keep')
    keep_table = s_table.get_individual_table()
    keep_table.to_csv(keep_file, index=False, header=False, sep="\t")

    snp_keepfile = osp.join(temp_dir, f"variants.keep")
    pd.DataFrame({'SNP': genotype_matrix.snps}).to_csv(
        snp_keepfile, index=False, header=False
    )

    plink_output = osp.join(temp_dir, "variants")

    cmd = [
        f"--bfile {genotype_matrix.bed_file}",
        f"--keep {keep_file}",
        f"--extract {snp_keepfile}",
        f"--missing variant-only",
        f"--out {plink_output}",
    ]

    plink2.execute(cmd)

    miss_df = pd.read_csv(plink_output + ".vmiss", delim_whitespace=True)
    miss_df = pd.DataFrame({'ID': genotype_matrix.snps}).merge(miss_df)

    if len(miss_df) != genotype_matrix.n_snps:
        raise ValueError("Length of missingness table does not match number of SNPs.")

    return (miss_df['OBS_CT'] - miss_df['MISSING_CT']).values




import numpy as np
from magenpy.simulation.GWASimulator import GWASimulator


class AnnotatedGWASimulator(GWASimulator):
    """
    Simulate complex traits by incorporating genomic functional
    annotations into the mixture densities that govern the effect size
    of each variant on the trait.

    NOTE: This code is experimental and needs much further validation.
    """

    def __init__(self, bed_files, **kwargs):
        super().__init__(bed_files, **kwargs)

        # For now, we will restrict to 2 mixture components.
        assert self.n_mixtures == 2

        self.w_h2 = None  # The annotation weights for the per-SNP heritability
        self.w_pi = None  # The annotation weights for the per-SNP causal probability

    def set_w_h2(self, w_h2):
        """
        Set the annotation weights for the per SNP heritability
        :param w_h2: A vector of weights for each annotation.
        """

        assert len(w_h2) == self.n_annotations

        self.w_h2 = w_h2
        self.set_per_snp_heritability()

    def simulate_w_h2(self, enrichment=None):
        """
        Simulate the annotation weights for the per-SNP heritability
        """
        raise NotImplementedError

    def set_w_pi(self, w_pi):
        """
        Set the annotation weights for the per SNP causal probability
        :param w_pi: A vector of weights for each annotation.
        """

        assert len(w_pi) == self.n_annotations

        self.w_pi = w_pi
        self.set_per_snp_mixture_probability()

    def simulate_w_pi(self, enrichment=None):
        """
        Simulate the annotation weights for the per-SNP causal probability

        :param enrichment: A dictionary of enrichment values where the
        key is the annotation and the value is the enrichment
        """

        enrichment = enrichment or {}
        enr = []
        for annot in self.annotation[self.chromosomes[0]].annotations:
            try:
                enr.append(enrichment[annot])
            except KeyError:
                enr.append(1.)

        self.w_pi = np.log(np.array(enr))

    def set_per_snp_heritability(self):

        if self.w_h2 is None:
            return super().set_per_snp_heritability()

        self.per_snp_h2 = {}

        for c in self.chromosomes:
            self.per_snp_h2[c] = np.clip(np.dot(self.annotation[c].values(), self.w_h2),
                                         a_min=0., a_max=np.inf)

    def set_per_snp_mixture_probability(self):

        if self.w_pi is None:
            return super().set_per_snp_mixture_probability()

        self.per_snp_pi = {}

        for c in self.chromosomes:
            prob = 1./(1. + np.exp(-np.dot(self.annotation[c].values(add_intercept=True),
                                           np.concatenate([[np.log(self.pi[1])], self.w_pi]))))
            self.per_snp_pi[c] = np.array([1. - prob, prob]).T

    def get_heritability_enrichment(self):

        tabs = self.to_true_beta_table(per_chromosome=True)
        total_heritability = sum([tab['Heritability'].sum() for c, tab in tabs.items()])

        heritability_per_binary_annot = {
            bin_annot: 0. for bin_annot in self.annotation[self.chromosomes[0]].binary_annotations
        }

        n_variants_per_binary_annot = {
            bin_annot: 0 for bin_annot in heritability_per_binary_annot
        }

        for c, c_size in self.shapes.items():
            for bin_annot in self.annotation[c].binary_annotations:
                annot_idx = self.annotation[c].get_binary_annotation_index(bin_annot)
                heritability_per_binary_annot[bin_annot] += tabs[c].iloc[np.array(annot_idx), :]['Heritability'].sum()
                n_variants_per_binary_annot[bin_annot] += len(annot_idx)

        return {
            bin_annot: (ba_h2/total_heritability) / (n_variants_per_binary_annot[bin_annot] / self.m)
            for bin_annot, ba_h2 in heritability_per_binary_annot.items()
        }




from typing import Union
import numpy as np
import pandas as pd
from .parsers.plink_parsers import parse_fam_file


class SampleTable(object):

    def __init__(self, table: Union[pd.DataFrame, None] = None, phenotype_likelihood: Union[str, None] = None):

        self.table: Union[pd.DataFrame, None] = table

        assert phenotype_likelihood in (None, 'binomial', 'gaussian')

        self._phenotype_likelihood: Union[str, None] = phenotype_likelihood
        self._covariate_cols = None

        self.post_check_phenotype()

    @property
    def shape(self):
        return (self.n,)

    @property
    def n(self):
        return len(self.table)

    @property
    def sample_size(self):
        return self.n

    @property
    def iid(self):
        if self.table is not None:
            return self.table['IID'].values

    @property
    def fid(self):
        if self.table is not None:
            return self.table['FID'].values

    @property
    def phenotype(self):
        if self.table is not None:
            try:
                return self.table['phenotype'].values
            except KeyError:
                raise KeyError("The phenotype is not set!")

    @property
    def covariates(self):
        return self._covariate_cols

    @property
    def phenotype_likelihood(self):
        return self._phenotype_likelihood

    @classmethod
    def from_fam_file(cls, fam_file):
        s_tab = parse_fam_file(fam_file)
        return cls(table=s_tab)

    @classmethod
    def from_phenotype_file(cls, phenotype_file, filter_na=True, **read_csv_kwargs):
        s_tab = cls()
        s_tab.read_phenotype_file(phenotype_file, filter_na, **read_csv_kwargs)
        return s_tab

    @classmethod
    def from_covariate_file(cls, covar_file, **read_csv_kwargs):
        s_tab = cls()
        s_tab.read_covariates_file(covar_file, **read_csv_kwargs)
        return s_tab

    def read_phenotype_file(self, phenotype_file, drop_na=True, **read_csv_kwargs):
        """
        Read the phenotype file from disk. The expected format is Family ID (`FID`),
        Individual ID (`IID`) and the phenotype column `phenotype`. You may adjust
        the parsing configurations with keyword arguments that will be passed to `pandas.read_csv`.

        :param phenotype_file: The path to the phenotype file.
        :param drop_na: Drop samples whose phenotype value is missing (Default: True).
        :param read_csv_kwargs: keyword arguments to pass to the `read_csv` function of `pandas`.
        """

        if 'sep' not in read_csv_kwargs and 'delimiter' not in read_csv_kwargs:
            read_csv_kwargs['delim_whitespace'] = True

        if 'na_values' not in read_csv_kwargs:
            read_csv_kwargs['na_values'] = {'phenotype': [-9.]}

        if 'dtype' not in read_csv_kwargs:
            read_csv_kwargs['dtype'] = {'phenotype': float}

        pheno_table = pd.read_csv(phenotype_file, **read_csv_kwargs)
        pheno_table.columns = ['FID', 'IID', 'phenotype']

        if self.table is not None:
            pheno_table['FID'] = pheno_table['FID'].astype(type(self.fid[0]))
            pheno_table['IID'] = pheno_table['IID'].astype(type(self.iid[0]))

            self.table = self.table.merge(pheno_table)
        else:
            self.table = pheno_table

        if self.table['phenotype'].isnull().all():
            self.table.drop('phenotype', axis=1, inplace=True)
        elif drop_na:
            # Maybe using converters in the read_csv above?
            self.table = self.table.dropna(subset=['phenotype'])

        self.post_check_phenotype()

    def read_covariates_file(self, covar_file, **read_csv_kwargs):

        """
        Read the covariates file from the provided path. The expected format is Family ID (`FID`),
        Individual ID (`IID`) and the remaining columns are assumed to be covariates. You may adjust
        the parsing configurations with keyword arguments that will be passed to `pandas.read_csv`.

        :param covar_file: The path to the covariates file.
        :param read_csv_kwargs: keyword arguments to pass to the `read_csv` function of `pandas`.
        """

        if 'sep' not in read_csv_kwargs and 'delimiter' not in read_csv_kwargs:
            read_csv_kwargs['delim_whitespace'] = True

        covar_table = pd.read_csv(covar_file, **read_csv_kwargs)
        self._covariate_cols = covar_table.columns[2:]
        covar_table.columns = ['FID', 'IID'] + self._covariate_cols

        if self.table is not None:
            covar_table['FID'] = covar_table['FID'].astype(type(self.fid[0]))
            covar_table['IID'] = covar_table['IID'].astype(type(self.iid[0]))

            self.table = self.table.merge(covar_table)
        else:
            self.table = covar_table

    def post_check_phenotype(self):
        """
        Apply some simple heuristics to check the phenotype values
        provided by the user and infer the phenotype likelihood (if needed).
        """

        if 'phenotype' in self.table.columns:

            unique_vals = self.table['phenotype'].unique()

            if self.table['phenotype'].isnull().all():
                self.table.drop('phenotype', axis=1, inplace=True)
            elif self.phenotype_likelihood in ('binomial', None):

                if len(unique_vals) > 2:
                    self._phenotype_likelihood = 'gaussian'
                    return

                unique_vals = sorted(unique_vals)
                if unique_vals == [1, 2]:
                    # Plink coding for case/control
                    self.table['phenotype'] -= 1
                    self._phenotype_likelihood = 'binomial'
                elif unique_vals == [0, 1]:
                    self._phenotype_likelihood = 'binomial'
                else:
                    raise ValueError(f"Unknown values for binary traits: {unique_vals}. "
                                     f"The software only supports 0/1 or 1/2 coding for cases and controls.")

    def filter_samples(self, keep_samples=None, keep_file=None):
        """
        Filter samples from the samples table. User must specify
        either a list of samples to keep or the path to a file
        with the list of samples to keep.

        :param keep_samples: A list (or array) of sample IDs to keep.
        :param keep_file: The path to a file with the list of samples to keep.
        """

        assert keep_samples is not None or keep_file is not None

        if keep_samples is None:
            from .parsers.misc_parsers import read_sample_filter_file
            keep_samples = read_sample_filter_file(keep_file)

        self.table = self.table.merge(pd.DataFrame({'IID': keep_samples},
                                                   dtype=type(self.iid[0])))

    def get_table(self, col_subset=None):
        if col_subset is not None:
            return self.table[list(col_subset)]
        else:
            return self.table

    def get_individual_table(self):
        return self.get_table(col_subset=['FID', 'IID'])

    def get_phenotype_table(self):
        try:
            return self.get_table(col_subset=['FID', 'IID', 'phenotype'])
        except KeyError:
            raise KeyError("The phenotype is not set!")

    def get_covariates_table(self, covar_subset=None):
        assert self._covariate_cols is not None

        if covar_subset is None:
            covar = self._covariate_cols
        else:
            covar = list(set(self._covariate_cols).intersection(set(covar_subset)))

        assert len(covar) >= 1

        return self.get_table(col_subset=['FID', 'IID'] + covar)

    def get_covariates(self, covar_subset=None):
        return self.get_covariates_table(covar_subset=covar_subset).iloc[:, 2:].values

    def set_phenotype(self, phenotype, phenotype_likelihood=None):

        self.table['phenotype'] = phenotype

        if phenotype_likelihood:
            self._phenotype_likelihood = phenotype_likelihood

    def to_file(self, output_file, col_subset=None, **to_csv_kwargs):
        """
        Write the sample table to file.
        :param output_file: The path to the file where to write the sample table.
        :param col_subset: A subset of the columns to write to file.
        """

        if 'sep' not in to_csv_kwargs and 'delimiter' not in to_csv_kwargs:
            to_csv_kwargs['sep'] = '\t'

        if 'index' not in to_csv_kwargs:
            to_csv_kwargs['index'] = False

        if col_subset is not None:
            table = self.table[col_subset]
        else:
            table = self.table

        table.to_csv(output_file, **to_csv_kwargs)

    def __len__(self):
        return self.n

    def __eq__(self, other):
        return np.array_equal(self.iid, other.iid)




def standardize(g_mat, fill_na=True):
    """
    Standardize the genotype matrix, such that the columns (i.e. snps)
    have zero mean and unit variance.
    :param g_mat: A two dimensional matrix (numpy, dask, xarray, etc.)
    :param fill_na: If true, fill the missing values with zero after standardizing.
    """
    sg_mat = (g_mat - g_mat.mean(axis=0)) / g_mat.std(axis=0)

    if fill_na:
        sg_mat = sg_mat.fillna(0.)

    return sg_mat



from .system_utils import is_cmd_tool, run_shell_script, available_cpu
import magenpy as mgp


class plink2Executor(object):

    def __init__(self, threads='auto'):

        if threads == 'auto':
            self.threads = available_cpu()
        else:
            self.threads = threads

        self.plink2_path = mgp.get_option('plink2_path')

        if not is_cmd_tool(self.plink2_path):
            raise Exception(f"Did not find the executable for plink2 at: {self.plink2_path}")

    def execute(self, cmd):

        cmd = [self.plink2_path] + cmd + [f'--threads {self.threads}']
        run_shell_script(" ".join(cmd))


class plink1Executor(object):

    def __init__(self, threads='auto'):

        if threads == 'auto':
            self.threads = available_cpu()
        else:
            self.threads = threads

        self.plink1_path = mgp.get_option('plink1.9_path')

        if not is_cmd_tool(self.plink1_path):
            raise Exception(f"Did not find the executable for plink at: {self.plink1_path}")

    def execute(self, cmd):

        cmd = [self.plink1_path] + cmd + [f'--threads {self.threads}']
        run_shell_script(" ".join(cmd))



import numpy as np


def adjust_for_covariates(phenotype, covariates):
    """
    This function takes a phenotype vector and a matrix of covariates
    and applies covariate correction on the phenotype. Concretely,
    this involves fitting a linear model where the response is the
    phenotype and the predictors are the covariates and then returning
    the residuals.
    :param phenotype: A vector of continuous or quantitative phenotypes.
    :param covariates: A matrix where each row corresponds to an individual
     and each column corresponds to a covariate (e.g. age, sex, PCs, etc.)
    """

    import statsmodels.api as sm

    return sm.OLS(phenotype, sm.add_constant(covariates)).fit().resid


def rint(phenotype, offset=3./8):
    """
    Apply Rank-based inverse normal transform on the phenotype.
    :param phenotype: A vector of continuous or quantitative phenotypes.
    :param offset: The offset to use in the INT transformation (Blom's offset by default).
    """

    from scipy.stats import rankdata, norm

    ranked_pheno = rankdata(phenotype, method="average")
    return norm.ppf((ranked_pheno - offset) / (len(ranked_pheno) - 2 * offset + 1))


def find_outliers(phenotype, sigma_threshold=5):
    """
    Detect samples with outlier phenotype values.
    This function takes a vector of quantitative phenotypes,
    computes the z-score for every individual, and returns a
    boolean vector indicating whether individual i has phenotype value
    within the specified standard deviations `sigma_threshold`.
    :param sigma_threshold: The multiple of standard deviations or sigmas after
    which we consider the phenotypic value an outlier.
    """
    from scipy.stats import zscore
    return np.abs(zscore(phenotype)) < sigma_threshold


def standardize(phenotype):
    """
    Standardize the phenotype vector to have mean zero and unit variance
    :param phenotype: A numpy vector of continuous or quantitative phenotypes.
    """
    return (phenotype - phenotype.mean()) / phenotype.std()



"""
Author: Shadi Zabad
Date: March 2021
"""

import pandas as pd
import numpy as np
from magenpy import GWADataLoader
from magenpy.simulation.GWASimulator import GWASimulator


class MulticohortGWASimulator(GWADataLoader):
    """
    A module for simulating GWAS data for separate cohorts.
    This includes scenarios such as multi-population or multi-ethnic datasets, or 
    datasets that can be stratified by a discrete variable.

    NOTE: Incomplete code.. requires more testing.

    """

    def __init__(self,
                 bed_files,
                 cluster_assignments_file,
                 prop_shared_causal=1.,
                 rho=1.,
                 **kwargs):
        """
        Simulate phenotypes using the linear additive model while accounting 
        for heterogeneous genetic architectures across cohorts.
    
        :param bed_files: A path (or list of paths) to PLINK BED files.
        :param cluster_assignments_file: A file mapping each sample in the BED files to their corresponding 
        cohort or cluster.
        :param prop_shared_causal: Proportion of causal variants that are shared across clusters.
        :param rho: The correlation coefficient for the effect size across clusters.
        """

        super().__init__(bed_files, **kwargs)

        from ..parsers.misc_parsers import parse_cluster_assignment_file

        self.cluster_table = parse_cluster_assignment_file(cluster_assignments_file)

        # Proportion of causal snps that are shared
        self.prop_shared_causal = prop_shared_causal

        # Rho can be either a scalar or a matrix that determines the patterns of
        # correlations between effect sizes in different clusters.
        if type(rho) == np.float:
            self.rho = rho*np.ones(shape=(len(self.clusters), len(self.clusters)))
            np.fill_diagonal(self.rho, 1.)
        else:
            self.rho = rho

        # Reference cluster
        self.ref_cluster = None

        # A dictionary of GWAS simulators for each cluster
        self.cluster_simulators = {}

        for c in self.clusters:
            if self.ref_cluster is None:
                self.ref_cluster = c

            self.cluster_simulators[c] = GWASimulator(bed_files,
                                                      keep_samples=self.get_samples_in_cluster(c),
                                                      **kwargs)

    @property
    def clusters(self):
        return self.cluster_table['Cluster'].unique()

    def get_samples_in_cluster(self, cluster):
        return self.cluster_table.loc[self.cluster_table['Cluster'] == cluster, 'IID'].values

    def set_reference_cluster(self, c):
        self.ref_cluster = c

    def simulate_causal_status(self):

        # The reference simulator:
        ref_sim = self.cluster_simulators[self.ref_cluster]

        # Simulate causal snps in reference cluster:
        ref_sim.simulate_mixture_assignment()

        # Get the causal snps in reference cluster:
        ref_causal = {
            c: np.where(a)[0]
            for c, a in ref_sim.get_causal_status().items()
        }

        for c in self.clusters:
            # For each cluster that is not the reference,
            # update their causal snps according to our draw for
            # the reference cluster
            if c != self.ref_cluster:

                new_mixture = ref_sim.mixture_assignment.copy()

                if self.prop_shared_causal < 1.:
                    for ch, ref_c in ref_causal.items():

                        # The number of shared causal snps for Chromosome `ch`:
                        n_shared_causal = int(np.floor(self.prop_shared_causal * len(ref_c)))

                        # Number of snps to flip:
                        n_flip = len(ref_c) - n_shared_causal

                        # Randomly decide which snps to "turn off":
                        for i in np.random.choice(ref_c, size=n_flip, replace=False):
                            new_mixture[ch][i] = new_mixture[ch][i][::-1]
                            # With probability p, switch on some other randomly chosen SNP:
                            # NOTE: If the number of SNPs is small, there's a small chance
                            # that this may flip the same SNP multiple times.
                            if np.random.uniform() < ref_sim.pis[1]:
                                new_i = np.random.choice(self.shapes[ch])
                                new_mixture[ch][new_i] = new_mixture[ch][new_i][::-1]

                self.cluster_simulators[c].set_mixture_assignment(
                    new_mixture
                )

    def simulate_beta(self):

        for c in self.clusters:
            self.cluster_simulators[c].beta = {}

        for ch, c_size in self.shapes.items():
            # Draw the beta from a multivariate normal distribution with covariance
            # as specified in the matrix `rho`.
            betas = np.random.multivariate_normal(np.zeros(self.rho.shape[0]), cov=self.rho, size=c_size)
            for i, c in enumerate(self.clusters):
                self.cluster_simulators[c].beta[ch] = (
                        self.cluster_simulators[c].get_causal_status()[ch].astype(np.int)*betas[:, i]
                )

    def simulate(self, perform_gwas=False):

        self.simulate_causal_status()
        self.simulate_beta()

        iids = self.sample_table.iid

        phenotypes = pd.Series(np.zeros_like(iids), index=iids)

        for c in self.clusters:
            self.cluster_simulators[c].simulate(reset_beta=False)
            phenotypes[self.cluster_simulators[c].sample_table.iid] = self.cluster_simulators[c].sample_table.phenotype

        self.set_phenotype(phenotypes)

        # Perform GWAS on the pooled sample:
        if perform_gwas:
            self.perform_gwas()



"""
Author: Shadi Zabad
Date: March 2021
"""

import warnings
import numpy as np
import pandas as pd
from magenpy.GWADataLoader import GWADataLoader


class GWASimulator(GWADataLoader):

    def __init__(self,
                 bed_files,
                 h2=0.2,
                 pi=(0.9, 0.1),
                 d=(0., 1.),
                 prevalence=0.15,
                 **kwargs):
        """
        Simulate phenotypes using the linear additive model.

        :param bed_files: A path (or list of paths) to PLINK BED files.
        :param h2: The trait SNP heritability, or proportion of variance explained by SNPs.
        :param pi: The mixture proportions for Gaussian mixture density.
        :param d:  The variance multipliers for each component of the Gaussian mixture density.
        :param prevalence: The (disease) prevalence for binary (case-control) phenotypes.
        """

        super().__init__(bed_files, **kwargs)

        self.pi = pi
        self.h2 = h2
        self.prevalence = prevalence

        # Sanity checks:
        assert 0. <= self.h2 <= 1.
        assert round(sum(self.pi), 1) == 1.
        assert 0. < self.prevalence < 1.

        self.d = np.array(d)

        self.per_snp_h2 = None
        self.per_snp_pi = None
        self.beta = None
        self.mixture_assignment = None

    @property
    def n_mixtures(self):
        return len(self.pi)

    def set_pi(self, new_pi):
        """
        Set the mixture proportions (proportion of variants in each
        Gaussian mixture component).
        """
        self.pi = new_pi
        self.set_per_snp_mixture_probability()

    def set_h2(self, new_h2):
        """
        Set the total heritability (proportion of additive variance due to SNPs) for the trait
        """
        self.h2 = new_h2
        self.set_per_snp_heritability()

    def set_per_snp_mixture_probability(self):
        """
        Set the per-SNP mixture probability for each variant in the dataset.
        """

        self.per_snp_pi = {}

        for c, c_size in self.shapes.items():
            self.per_snp_pi[c] = np.repeat(np.array([self.pi]), c_size, axis=0)

    def set_per_snp_heritability(self):
        """
        Set the per-SNP heritability for each variant in the dataset.
        """

        assert self.mixture_assignment is not None

        # Estimate the global sigma_beta_sq based on the
        # pre-specified heritability, the mixture proportions `pi`,
        # and the prior multipliers `d`.

        combined_assignments = np.concatenate([self.mixture_assignment[c] for c in self.chromosomes])

        sigma_beta_sq = self.h2 / (combined_assignments*self.d).sum()

        self.per_snp_h2 = {}

        for c, c_size in self.shapes.items():
            self.per_snp_h2[c] = sigma_beta_sq*self.d[np.where(self.mixture_assignment[c])[1]]

    def get_causal_status(self):
        """
        This method returns a dictionary where the keys are the chromosome numbers
        and the values are of binary vectors indicating which SNPs are causal for the phenotype.
        """

        assert self.mixture_assignment is not None

        try:
            zero_index = list(self.d).index(0)
        except ValueError:
            # If all SNPs are causal:
            return {c: np.repeat(True, c_size) for c, c_size in self.shapes.items()}

        causal_status = {}

        for c, mix_a in self.mixture_assignment.items():
            causal_status[c] = np.where(mix_a)[1] != zero_index

        return causal_status

    def set_causal_snps(self, causal_snps):
        """
        A utility method to set the causal SNPs in the simulation based on an array or
        list of SNPs specified by the user. The method takes an iterable of `causal_snps`
        and then creates a new mixture assignment object where only the `causal_snps`
        contribute to the phenotype.

        :param causal_snps: A list or array of SNP IDs.
        """

        # Get the index of the mixture component whose multiplier is zero (i.e. the null component):
        try:
            zero_index = list(self.d).index(0)
        except ValueError:
            raise ValueError("Cannot set causal variants when all mixture components are causal. Modify "
                             "the mixture multipliers `d` to enable this functionality.")

        # Get the indices of the non-null mixture components:
        nonzero_indices = [i for i, d in enumerate(self.d) if d != 0.]

        # Get the mixture proportions for the non-null components and normalize them so they sum to 1:
        pis = np.array(self.pi)[nonzero_indices]
        pis /= pis.sum()

        # Initialize new mixture assignment object:
        new_assignment = {c: np.zeros((s, self.n_mixtures)) for c, s in self.shapes.items()}

        from magenpy.utils.compute_utils import intersect_arrays

        n_causal_set = 0

        for c, snps in self.snps.items():

            # Intersect the list of causal SNPs with the SNPs on chromosome `c`:
            snp_idx = intersect_arrays(snps, causal_snps, return_index=True)

            if len(snp_idx) > 0:
                n_causal_set += len(snp_idx)
                # For the causal SNPs, assign them randomly to the causal components
                new_assignment[c][snp_idx, np.random.choice(nonzero_indices,
                                                            size=len(snp_idx),
                                                            p=pis)] = 1
                # For the remaining SNPs, assign them to the null component:
                new_assignment[c][:, zero_index] = np.abs(new_assignment[c].sum(axis=1) - 1)

        if n_causal_set < len(causal_snps):
            warnings.warn("Not all causal SNPs are represented in the genotype matrix! "
                          f"User passed a list of {len(causal_snps)} SNPs, only matched {n_causal_set}.")

        self.set_mixture_assignment(new_assignment)

    def set_mixture_assignment(self, new_assignment):
        """
        Set the mixture assignments according to user-provided dictionary.
        :param new_assignment: A dictionary where the keys are the chromosomes and
        the values are the mixture assignment for each SNP on that chromosome.
        """

        # Check that the shapes match pre-specified information:
        for c, c_size in self.shapes.items():
            assert new_assignment[c].shape == (c_size, self.n_mixtures)

        self.mixture_assignment = new_assignment

    def simulate_mixture_assignment(self):
        """
        Simulate assigning SNPs to the various mixture components
        with probabilities given by `pi`.
        """

        if self.per_snp_pi is None or len(self.per_snp_pi) < 1:
            self.set_per_snp_mixture_probability()

        self.mixture_assignment = {}

        from magenpy.utils.model_utils import multinomial_rvs

        for c, c_size in self.shapes.items():

            self.mixture_assignment[c] = multinomial_rvs(1, self.per_snp_pi[c])

        return self.mixture_assignment

    def set_beta(self, new_beta):
        """
        Set the beta according to user-provided dictionary.
        :param new_beta: A dictionary where the keys are the chromosomes and
        the values are the beta for each SNP on that chromosome.
        """

        # Check that the shapes match pre-specified information:
        for c, c_size in self.shapes.items():
            assert len(new_beta[c]) == c_size

        self.beta = new_beta

    def simulate_beta(self):
        """
        Simulate the causal effect size for the variants included
        in the dataset.
        """

        if self.per_snp_h2 is None or len(self.per_snp_h2) < 1:
            self.set_per_snp_heritability()

        self.beta = {}

        for c, c_size in self.shapes.items():

            self.beta[c] = np.random.normal(loc=0.0,
                                            scale=np.sqrt(self.per_snp_h2[c]),
                                            size=c_size)

        return self.beta

    def simulate_phenotypes(self):
        """
        Simulate complex phenotypes for the samples, given their genotype information and 
        fixed effect sizes `beta` that were simulated previously.
        """

        assert self.beta is not None

        # Compute the polygenic score given the simulated/provided beta:
        pgs = self.score(self.beta)

        # Sample the environmental/residual component:
        e = np.random.normal(loc=0., scale=np.sqrt(1. - self.h2), size=self.sample_size)

        # The final simulated phenotype is a combination of
        # the polygenic score + the residual component:
        y = pgs + e

        if self.phenotype_likelihood == 'binomial':
            # If the simulated phenotype is to be binary,
            # use the threshold model to determine positives/negatives
            # based on the prevalence of the phenotype in the population:
            
            from ..stats.transforms.phenotype import standardize
            
            y = standardize(y)

            from scipy.stats import norm
            
            cutoff = norm.ppf(1. - self.prevalence)
            new_y = np.zeros_like(y, dtype=int)
            new_y[y > cutoff] = 1
        else:
            new_y = y

        self.set_phenotype(new_y)
        
        return new_y

    def simulate(self, reset_beta=True, reset_mixture_assignment=True, perform_gwas=False):
        """
        A convenience method to simulate all the components of the generative model.
        Specifically, the simulation follows the standard linear model, where the phenotype is 
        dependent on the genotype + environmental components that are assumed to be uncorrelated:
        
        `Y = XB + e`
        
        Where `Y` is the vector of phenotypes, `X` is the genotype matrix, `B` is the vector of effect sizes, 
        and `e` represents the residual effects. The generative model proceeds by:
         
         1) Drawing the effect sizes `beta` from a Gaussian mixture density.
         2) Drawing the residual effect from an isotropic Gaussian density.
         3) Setting the phenotype according to the equation above. 
        
        :param reset_beta: If True, reset the effect sizes by drawing new ones from the prior density.
        :param reset_mixture_assignment: If True, reset the assignment of SNPs to mixture components. Set to False
        if you'd like to keep the same configuration of causal SNPs.
        :param perform_gwas: If True, automatically perform genome-wide association on the newly simulated phenotype.
        """

        # Simulate the mixture assignment:
        if self.mixture_assignment is None or reset_mixture_assignment:
            self.simulate_mixture_assignment()

        # Set the per-SNP heritability based on the mixture assignment:
        self.set_per_snp_heritability()

        # Simulate betas based on per-SNP heritability
        if self.beta is None or reset_beta:
            self.simulate_beta()

        # Simulate the phenotype
        self.simulate_phenotypes()

        if perform_gwas:
            # Perform genome-wide association testing...
            self.perform_gwas()

    def to_true_beta_table(self, per_chromosome=False):
        """
        Export the simulated true effect sizes and causal status
         into a pandas table.
        :param per_chromosome: If True, return a dictionary of tables for each chromosome.
        """

        assert self.beta is not None

        eff_tables = {}
        causal_status = self.get_causal_status()

        for c in self.chromosomes:

            eff_tables[c] = pd.DataFrame({
                'CHR': c,
                'SNP': self.genotype[c].snps,
                'A1': self.genotype[c].a1,
                'MixtureComponent': np.where(self.mixture_assignment[c] == 1)[1],
                'Heritability': self.per_snp_h2[c],
                'BETA': self.beta[c].flatten(),
                'Causal': causal_status[c],
            })

        if per_chromosome:
            return eff_tables
        else:
            return pd.concat(list(eff_tables.values()))



import numpy as np
import magenpy as mgp


def simple_ldsc(gdl: mgp.GWADataLoader):
    """
    Provides an estimate of SNP heritability from summary statistics using
    a simplified version of the LD Score Regression framework.
    E[X_j^2] = h^2*l_j + int
    Where the response is the Chi-Squared statistic for SNP j
    and the variable is its LD score.

    NOTE: For now, we constrain the slope to 1.

    :param gdl: An instance of `GWADataLoader` with the LD information and
    summary statistics initialized properly.
    """

    # Check data types:
    assert gdl.ld is not None and gdl.sumstats_table is not None

    ld_score = []
    chi_sq = []
    sample_size = []

    for c in gdl.chromosomes:
        ld_score.append(gdl.ld[c].ld_score)
        chi_sq.append(gdl.sumstats_table[c].get_chisq_statistic())
        sample_size.append(gdl.sumstats_table[c].n_per_snp.max())

    ld_score = np.concatenate(ld_score)
    chi_sq = np.concatenate(chi_sq)
    sample_size = max(sample_size)

    return (chi_sq.mean() - 1.) * len(ld_score) / (ld_score.mean() * sample_size)


class LDSCRegression(object):

    def __init__(self, gdl: mgp.GWADataLoader, n_blocks=200, max_chisq=None):
        """
        Incomplete...
        """

        self.gdl = gdl
        self.n_blocks = n_blocks

        # Extract the data from the GDL object:

        chroms = self.gdl.chromosomes

        if self.gdl.annotation is not None:
            self.ld_scores = np.concatenate([
                self.gdl.ld[c].compute_ld_scores(
                    annotations=self.gdl.annotation[c].values(add_intercept=True)
                )
                for c in chroms
            ])
        else:
            self.ld_scores = np.concatenate([self.gdl.ld[c].ld_score.reshape(-1, 1) for c in chroms])

        self.chisq = np.concatenate([self.gdl.sumstats_table[c].get_chisq_statistic() for c in chroms])
        self.n = np.concatenate([self.gdl.sumstats_table[c].n_per_snp for c in chroms])

        if max_chisq is None:
            max_chisq = max(0.001*self.n.max(), 80)

        chisq_cond = self.chisq < max_chisq

        self.ld_scores = self.ld_scores[chisq_cond, :]
        self.chisq = self.chisq[chisq_cond]
        self.n = self.n[chisq_cond]

    def fit(self):
        """
        TODO: Implement the jackknife estimator here...
        """
        pass




import glob
import configparser

# Data structures:
from magenpy.AnnotationMatrix import AnnotationMatrix
from magenpy.LDMatrix import LDMatrix
from magenpy.GWADataLoader import GWADataLoader
from magenpy.SumstatsTable import SumstatsTable
from magenpy.SampleTable import SampleTable

# Simulation:

from magenpy.simulation.GWASimulator import GWASimulator

# Data utilities:

from magenpy.utils.data_utils import *

__version__ = '0.0.10'


config = configparser.ConfigParser()
config.read(glob.glob(osp.join(osp.dirname(__file__), 'config/*.ini')))


def print_options():
    """
    Print the options stored in the configuration file
    """
    for sec in config.sections() + [config.default_section]:
        print("-> Section:", sec)
        for key in config[sec]:
            print(f"---> {key}: {config[sec][key]}")


def get_option(key):
    """
    Get the option associated with a given key
    """
    try:
        return config['USER'][key]
    except KeyError:
        return config['DEFAULT'][key]


def set_option(key, value):
    """
    Set an option in the configuration file by providing a key and a value
    """
    if 'USER' in config.sections():
        config['USER'][key] = value
    else:
        config['USER'] = {key: value}

    with open(osp.join(osp.dirname(__file__), 'config/paths.ini'), 'w') as configfile:
        config.write(configfile)



from tqdm import tqdm
import numpy as np
from scipy import stats


def merge_snp_tables(ref_table,
                     alt_table,
                     how='inner',
                     signed_statistics=('BETA', 'STD_BETA', 'Z'),
                     drop_duplicates=True,
                     correct_flips=True):
    """
    This function takes a reference SNP table with at least 3 columns ('SNP', 'A1', `A2`)
    and matches it with an alternative table that also has these 3 columns defined.
    The manner in which the join operation takes place depends on the `how` argument.
    Currently, the function supports `inner` and `left` joins.

    The function removes duplicates if `drop_dupicates` parameter is set to True

    If `correct_flips` is set to True, the function will correct summary statistics in
    the alternative table (e.g. BETA, MAF) based whether the A1 alleles agree between the two tables.


    :param ref_table: The reference table (pandas dataframe).
    :param alt_table: The alternative table (pandas dataframe)
    :param how: `inner` or `left`
    :param signed_statistics: The columns with signed statistics to flip if `correct_flips` is set to True.
    :param drop_duplicates: Drop duplicate SNPs
    :param correct_flips: Correct SNP summary statistics that depend on status of alternative allele
    """

    assert how in ('left', 'inner')

    merged_table = ref_table[['SNP', 'A1', 'A2']].merge(alt_table, how=how, on='SNP')

    if drop_duplicates:
        merged_table.drop_duplicates(inplace=True, subset=['SNP'])

    if how == 'left':
        merged_table['A1_y'] = merged_table['A1_y'].fillna(merged_table['A1_x'])
        merged_table['A2_y'] = merged_table['A2_y'].fillna(merged_table['A2_x'])

    # Assign A1 to be the one derived from the reference table:
    merged_table['A1'] = merged_table['A1_x']
    merged_table['A2'] = merged_table['A2_x']

    # Detect cases where the correct allele is specified in both tables:
    matching_allele = np.all(merged_table[['A1_x', 'A2_x']].values == merged_table[['A1_y', 'A2_y']].values, axis=1)

    # Detect cases where the effect and reference alleles are flipped:
    flip = np.all(merged_table[['A2_x', 'A1_x']].values == merged_table[['A1_y', 'A2_y']].values, axis=1)

    # Variants to keep:
    keep_snps = matching_allele | flip

    # Keep only SNPs with matching alleles or SNPs with flipped alleles:
    merged_table = merged_table.loc[keep_snps, ]

    if correct_flips:

        flip = flip[keep_snps].astype(int)
        num_flips = flip.sum()

        if num_flips > 0:

            # If the user provided a single signed statistic as a string, convert to list first:
            if isinstance(signed_statistics, str):
                signed_statistics = [signed_statistics]

            # Loop over the signed statistics and correct them:
            for s_stat in signed_statistics:
                if s_stat in merged_table:
                    merged_table[s_stat] = (-2. * flip + 1.) * merged_table[s_stat]

            # Correct MAF:
            if 'MAF' in merged_table:
                merged_table['MAF'] = np.abs(flip - merged_table['MAF'])

    merged_table.drop(['A1_x', 'A1_y', 'A2_x', 'A2_y'], axis=1, inplace=True)

    return merged_table


def identify_mismatched_snps(gdl,
                             chrom=None,
                             n_iter=10,
                             G=100,
                             p_dentist_threshold=5e-8,
                             p_gwas_threshold=1e-2,
                             rsq_threshold=.95,
                             max_removed_per_iter=.005):
    """
    This function implements a simple quality control procedures
    that checks that the GWAS summary statistics (Z-scores)
    are consistent with the LD reference panel. This is done
    using a simplified version of the framework outlined in the DENTIST paper:

    Improved analyses of GWAS summary statistics by reducing data heterogeneity and errors
    Chen et al. 2021

    Compared to DENTIST, the simplifications we make are:
        -   For each SNP, we sample one neighboring SNP at a time and compute the T statistic
            using that neighbor's information. The benefit of this is that we don't need to
            invert any matrices, so it's a fast operation to run.
        -   To arrive at a more robust estimate, we sample up to `k` neighbors and average
            the T-statistic across those `k` neighbors.

    NOTE: May need to re-implement this to apply some of the constraints genome-wide
    rather than on a per-chromosome basis.

    :param gdl: A `GWADataLoader` object
    :param chrom: A chromosome
    :param n_iter: Number of iterations
    :param G: The number of neighboring SNPs to sample (default: 100)
    :param p_dentist_threshold: The Bonferroni-corrected P-value threshold (default: 5e-8)
    :param p_gwas_threshold: The nominal GWAS P-value threshold for partitioning variants (default: 1e-2)
    :param rsq_threshold: The R^2 threshold to select neighbors (neighbor's squared
    correlation coefficient must be less than specified threshold).
    :param max_removed_per_iter: The maximum proportion of variants removed in each iteration
    """

    if chrom is None:
        chromosomes = gdl.chromosomes
    else:
        chromosomes = [chrom]

    shapes = gdl.shapes
    mismatched_dict = {c: np.repeat(False, c_size)
                       for c, c_size in gdl.shapes.items()}

    p_gwas_above_thres = {c: p_val > p_gwas_threshold for c, p_val in gdl.p_values.items()}
    gwas_thres_size = {c: p.sum() for c, p in p_gwas_above_thres.items()}
    converged = {c: False for c in gdl.chromosomes}

    for j in tqdm(range(n_iter),
                  total=n_iter,
                  desc="Identifying mismatched SNPs..."):

        for chrom in chromosomes:

            if converged[chrom]:
                continue

            ld_bounds = gdl.ld[chrom].get_masked_boundaries()
            z = gdl.z_scores[chrom]  # Obtain the z-scores
            t = np.zeros_like(z)

            # Loop over the LD matrix:
            for i, r in enumerate(gdl.ld[chrom]):

                # If the number of neighbors is less than 10, skip...
                if mismatched_dict[chrom][i] or len(r) < 10:
                    continue

                start_idx = ld_bounds[0, i]
                # Select neighbors randomly
                # Note: We are excluding neighbors whose squared correlation coefficient
                # is greater than pre-specified threshold:
                p = (np.array(r)**2 < rsq_threshold).astype(float)
                p /= p.sum()

                neighbor_idx = np.random.choice(len(r), p=p, size=G)
                neighbor_r = np.array(r)[neighbor_idx]

                # Predict the z-score of snp i, given the z-scores of its neighbors:
                pred_z = neighbor_r*z[start_idx + neighbor_idx]

                # Compute the Td statistic for each neighbor and average:
                t[i] = ((z[i] - pred_z) ** 2 / (1. - neighbor_r**2)).mean()

            # Compute the DENTIST p-value assuming a Chi-Square distribution with 1 dof.
            dentist_pval = 1. - stats.chi2.cdf(t, 1)
            # Use a Bonferroni correction to select mismatched SNPs:
            mismatched_snps = dentist_pval < p_dentist_threshold

            if mismatched_snps.sum() < 1:
                # If no new mismatched SNPs are identified, stop iterating...
                converged[chrom] = True
            elif j == n_iter - 1:
                # If this is the last iteration, take all identified SNPs
                mismatched_dict[chrom] = (mismatched_dict[chrom] | mismatched_snps)
            else:

                # Otherwise, we will perform the iterative filtering procedure
                # by splitting variants based on their GWAS p-values:

                # (1) Group S1: SNPs to remove from P_GWAS > threshold:
                mismatch_above_thres = mismatched_snps & p_gwas_above_thres[chrom]
                n_mismatch_above_thres = mismatch_above_thres.sum()
                prop_mismatch_above_thres = n_mismatch_above_thres / gwas_thres_size[chrom]

                if n_mismatch_above_thres < 1:
                    # If no mismatches are detected above the threshold, filter
                    # the mismatches below the threshold and continue...
                    mismatched_dict[chrom] = (mismatched_dict[chrom] | mismatched_snps)
                    continue

                # Sort the DENTIST p-values by index:
                sort_d_pval_idx = np.argsort(dentist_pval)

                if prop_mismatch_above_thres > max_removed_per_iter:
                    idx_to_keep = sort_d_pval_idx[mismatch_above_thres][
                                  int(gwas_thres_size[chrom]*max_removed_per_iter):]
                    mismatch_above_thres[idx_to_keep] = False

                # (2) Group S2: SNPs to remove from P_GWAS < threshold

                # Find mismatched variants below the threshold:
                mismatch_below_thres = mismatched_snps & (~p_gwas_above_thres[chrom])
                n_mismatch_below_thres = mismatch_below_thres.sum()
                prop_mismatch_below_thres = n_mismatch_below_thres / (shapes[chrom] - gwas_thres_size[chrom])

                # For the mismatched variants below the threshold,
                # we remove the same proportion as the variants above the threshold:
                prop_keep_below_thres = min(max_removed_per_iter, prop_mismatch_above_thres)

                if prop_mismatch_below_thres > prop_keep_below_thres:
                    idx_to_keep = sort_d_pval_idx[mismatch_below_thres][
                                  int((shapes[chrom] - gwas_thres_size[chrom]) * prop_keep_below_thres):
                                  ]
                    mismatch_below_thres[idx_to_keep] = False

                # Update the number of variants above the threshold:
                gwas_thres_size[chrom] -= mismatch_above_thres.sum()

                # Update the mismatched dictionary:
                mismatched_dict[chrom] = (mismatched_dict[chrom] | mismatch_below_thres | mismatch_above_thres)

    return mismatched_dict


def get_shared_distance_matrix(tree, tips=None):
    """
    This function takes a Biopython tree and returns the
    shared distance matrix (time to most recent common ancestor - MRCA)
    """

    tips = tree.get_terminals() if tips is None else tips
    n_tips = len(tips)  # Number of terminal species
    sdist_matrix = np.zeros((n_tips, n_tips))  # Shared distance matrix

    for i in range(n_tips):
        for j in range(i, n_tips):
            if i == j:
                sdist_matrix[i, j] = tree.distance(tree.root, tips[i])
            else:
                mrca = tree.common_ancestor(tips[i], tips[j])
                sdist_matrix[i, j] = sdist_matrix[j, i] = tree.distance(tree.root, mrca)

    return sdist_matrix


def tree_to_rho(tree, min_corr):
    """
    This function takes a Biopython tree and a minimum correlation
    parameter and returns the correlation matrix for the effect sizes
    across populations.

    :param tree: a Biopython Phylo object
    :param min_corr: minimum correlation
    :return:
    """

    max_depth = max(tree.depths().values())
    tree.root.branch_length = min_corr*max_depth / (1. - min_corr)
    max_depth = max(tree.depths().values())

    for c in tree.find_clades():
        c.branch_length /= max_depth

    return tree.root.branch_length + get_shared_distance_matrix(tree)


def multinomial_rvs(n, p):
    """
    Copied from Warren Weckesser:
    https://stackoverflow.com/a/55830796

    Sample from the multinomial distribution with multiple p vectors.

    * n must be a scalar.
    * p must an n-dimensional numpy array, n >= 1.  The last axis of p
      holds the sequence of probabilities for a multinomial distribution.

    The return value has the same shape as p.
    """
    count = np.full(p.shape[:-1], n)
    out = np.zeros(p.shape, dtype=int)
    ps = p.cumsum(axis=-1)
    # Conditional probabilities
    with np.errstate(divide='ignore', invalid='ignore'):
        condp = p / ps
    condp[np.isnan(condp)] = 0.0
    for i in range(p.shape[-1]-1, 0, -1):
        binsample = np.random.binomial(count, condp[..., i])
        out[..., i] = binsample
        count -= binsample
    out[..., 0] = count
    return out



import numpy as np
from magenpy import LDMatrix


class SampleLD(object):
    """
    Compute the sample correlation (LD) matrix between
    pairs of variants along a given chromosome.
    """

    def __init__(self, genotype_matrix):
        """
        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        """

        self.genotype_matrix = genotype_matrix

        # Ensure that the genotype matrix has data for a single chromosome only:
        if self.genotype_matrix.chromosome is None:
            raise Exception("We do not support computing inter-chromosomal LD matrices! "
                            "You may need to split the genotype matrix by chromosome. "
                            "See GenotypeMatrix.split_by_chromosome().")

    def compute_ld_boundaries(self):
        """
         Compute the Linkage-Disequilibrium (LD) boundaries. LD boundaries define the window
         for which we compute the correlation coefficient between the focal SNP and all other SNPs in
         the genome. Typically, this window is local, since the LD decays exponentially with
         genomic distance.

         The LD boundaries are a 2xM matrix, where M is the number of SNPs on the chromosome.
         The first row contains the start position for the window and the second row contains
         the end position.

         For the sample LD matrix, we simply take the entire square matrix as our window,
         so the start position is 0 and end position is M for all SNPs.
        """
        m = self.genotype_matrix.n_snps
        return np.array((np.zeros(m), np.ones(m)*m)).astype(np.int64)

    def compute(self, output_dir, temp_dir='temp'):
        """
        A utility method to compute the LD matrix and store in Zarr array format.
        The computes the LD matrix and stores it in Zarr array format, set its attributes,
        and performs simple validation at the end.

        :param output_dir: The path where to store the resulting LD matrix.
        :param temp_dir: A temporary directory to store intermediate files and results.
        """

        from .utils import compute_ld_xarray, compute_ld_plink1p9, _validate_ld_matrix
        from magenpy.GenotypeMatrix import xarrayGenotypeMatrix, plinkBEDGenotypeMatrix

        ld_boundaries = self.compute_ld_boundaries()

        if isinstance(self.genotype_matrix, xarrayGenotypeMatrix):
            zarr_ld_mat = compute_ld_xarray(self.genotype_matrix,
                                            ld_boundaries,
                                            output_dir,
                                            temp_dir=temp_dir)
        elif isinstance(self.genotype_matrix, plinkBEDGenotypeMatrix):
            zarr_ld_mat = compute_ld_plink1p9(self.genotype_matrix,
                                              ld_boundaries,
                                              output_dir,
                                              temp_dir=temp_dir)
        else:
            raise NotImplementedError

        zarr_ld_mat.attrs['Chromosome'] = int(self.genotype_matrix.chromosome)
        zarr_ld_mat.attrs['Sample size'] = int(self.genotype_matrix.sample_size)

        zarr_ld_mat.attrs['SNP'] = list(self.genotype_matrix.snps)
        zarr_ld_mat.attrs['BP'] = list(map(int, self.genotype_matrix.bp_pos))

        try:
            zarr_ld_mat.attrs['cM'] = list(map(float, self.genotype_matrix.cm_pos))
        except KeyError:
            pass

        zarr_ld_mat.attrs['MAF'] = list(map(float, self.genotype_matrix.maf))
        zarr_ld_mat.attrs['A1'] = list(self.genotype_matrix.a1)
        zarr_ld_mat.attrs['A2'] = list(self.genotype_matrix.a2)

        zarr_ld_mat.attrs['LD estimator'] = 'sample'
        zarr_ld_mat.attrs['LD boundaries'] = ld_boundaries.tolist()

        ld_mat = LDMatrix(zarr_ld_mat)
        ld_mat.set_store_attr('LDScore', ld_mat.compute_ld_scores().tolist())

        if _validate_ld_matrix(ld_mat):
            return ld_mat


class WindowedLD(SampleLD):
    """
    Compute the sample correlation matrix, but only in
    pre-specified windows for each variant.
    """

    def __init__(self,
                 genotype_matrix,
                 window_size=None,
                 kb_window_size=None,
                 cm_window_size=None):
        """
        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        :param window_size: The number of neighboring SNPs to consider on each side when computing LD.
        :param kb_window_size: The maximum distance in kilobases to consider when computing LD.
        :param cm_window_size: The maximum distance in centi Morgan to consider when computing LD.
        """

        super().__init__(genotype_matrix=genotype_matrix)

        assert not all([w is None for w in (window_size, kb_window_size, cm_window_size)])

        self.window_size = window_size
        self.kb_window_size = kb_window_size
        self.cm_window_size = cm_window_size

    def compute_ld_boundaries(self):
        """
         Compute the windowed Linkage-Disequilibrium (LD) boundaries.
         The LD boundaries computed here are the intersection of the windows defined by
         the window size around each SNP (`window_size`), the window size in kilobases (`kb_window_size`),
         and the window size in centi Morgan (`cm_window_size`).
        """

        bounds = []

        m = self.genotype_matrix.n_snps
        indices = np.arange(m)

        if self.window_size is not None:
            bounds.append(
                np.clip(np.array(
                    [indices - self.window_size,
                     indices + self.window_size
                     ]
                ),  a_min=0, a_max=m)
            )

        from .c_utils import find_windowed_ld_boundaries

        if self.kb_window_size is not None:
            bounds.append(
                find_windowed_ld_boundaries(.001*self.genotype_matrix.bp_pos,
                                            self.kb_window_size)
            )

        if self.cm_window_size is not None:
            bounds.append(
                find_windowed_ld_boundaries(self.genotype_matrix.cm_pos,
                                            self.cm_window_size)
            )

        if len(bounds) == 1:
            return bounds[0]
        else:
            return np.array([
                np.maximum.reduce([b[0, :] for b in bounds]),
                np.minimum.reduce([b[1, :] for b in bounds])
            ])

    def compute(self, output_dir, temp_dir='temp'):
        """
        :param output_dir: The path where to store the resulting LD matrix.
        :param temp_dir: A temporary directory to store intermediate files and results.
        """

        ld_mat = super(WindowedLD, self).compute(output_dir,
                                                 temp_dir)
        ld_mat.set_store_attr('LD estimator', 'windowed')

        w_properties = {}
        if self.window_size is not None:
            w_properties['Window size'] = self.window_size

        if self.kb_window_size is not None:
            w_properties['Window size (kb)'] = self.kb_window_size

        if self.cm_window_size is not None:
            w_properties['Window size (cM)'] = self.cm_window_size

        ld_mat.set_store_attr('Estimator properties', w_properties)

        return ld_mat


class ShrinkageLD(SampleLD):

    def __init__(self,
                 genotype_matrix,
                 genetic_map_ne,
                 genetic_map_sample_size,
                 threshold=1e-3):
        """
        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        :param genetic_map_ne: The effective population size (Ne) from which the genetic map is derived.
        :param genetic_map_sample_size: The sample size used to infer the genetic map.
        :param threshold: The shrinkage cutoff below which the LD is set to zero.
        """

        super().__init__(genotype_matrix=genotype_matrix)

        self.genetic_map_ne = genetic_map_ne
        self.genetic_map_sample_size = genetic_map_sample_size
        self.threshold = threshold

    def compute_ld_boundaries(self):
        """
        Find the LD boundaries based on the shrinkage operator.
        """

        from .c_utils import find_shrinkage_ld_boundaries
        return find_shrinkage_ld_boundaries(self.genotype_matrix.cm_pos,
                                            self.genetic_map_ne,
                                            self.genetic_map_sample_size,
                                            self.threshold)

    def compute(self, output_dir, temp_dir='temp'):
        """
        :param output_dir: The path where to store the resulting LD matrix.
        :param temp_dir: A temporary directory to store intermediate files and results.
        """

        ld_mat = super(ShrinkageLD, self).compute(output_dir,
                                                  temp_dir)

        from .utils import shrink_ld_matrix

        ld_mat = LDMatrix(shrink_ld_matrix(ld_mat.z_array,
                                           self.genotype_matrix.cm_pos,
                                           self.genotype_matrix.maf_var,
                                           self.genetic_map_ne,
                                           self.genetic_map_sample_size,
                                           self.threshold,
                                           ld_boundaries=ld_mat.ld_boundaries))

        ld_mat.set_store_attr('LD estimator', 'shrinkage')

        ld_mat.set_store_attr('Estimator properties', {
                    'Genetic map Ne': self.genetic_map_ne,
                    'Genetic map sample size': self.genetic_map_sample_size,
                    'Threshold': self.threshold
                })

        return ld_mat


class BlockLD(SampleLD):

    def __init__(self,
                 genotype_matrix,
                 ld_blocks=None,
                 ld_blocks_file=None):
        """
        :param genotype_matrix: The genotype matrix, an instance of `GenotypeMatrix`.
        :param ld_blocks: The LD blocks, a Bx2 matrix where B is the number of blocks and the
        columns are the start and end of each block, respectively.
        :param ld_blocks_file: The path to the LD blocks file
        """

        assert ld_blocks_file is not None or ld_blocks is not None

        super().__init__(genotype_matrix=genotype_matrix)

        if ld_blocks is None:
            from magenpy.parsers.misc_parsers import parse_ld_block_data
            self.ld_blocks = parse_ld_block_data(ld_blocks_file)[self.genotype_matrix.chromosome]

    def compute_ld_boundaries(self):
        """
        Find the per-SNP ld boundaries, given the provided LD blocks.
        """

        from .c_utils import find_ld_block_boundaries
        return find_ld_block_boundaries(self.genotype_matrix.bp_pos, self.ld_blocks)

    def compute(self, output_dir, temp_dir='temp'):
        """
        :param output_dir: The path where to store the resulting LD matrix.
        :param temp_dir: A temporary directory to store intermediate files and results.
        """

        ld_mat = super(BlockLD, self).compute(output_dir,
                                              temp_dir)

        ld_mat.set_store_attr('LD estimator', 'block')

        ld_mat.set_store_attr('Estimator properties', {
            'LD blocks': self.ld_blocks.tolist()
        })

        return ld_mat



import numpy as np
import pandas as pd
from magenpy.utils.compute_utils import intersect_arrays


class AnnotationMatrix(object):

    def __init__(self):

        self._annot_matrix = None
        self._snps = None
        self._chromosome = None
        self._annotations = None
        self._binary_annotations = None

    @classmethod
    def from_file(cls, annot_file, **read_csv_kwargs):
        """
        Takes an LDSC-style annotation file `annot_file` and reads the table
        into a pandas dataframe.
        NOTE: Assumes that SNPs are sorted by BP within a chromosome.

        :param annot_file: LDSC-style annotation file, containing columns for
        the chromosome (CHR), SNP ID (SNP), Base pair position (BP), position in centi Morgans (CM),
        and the annotations to load into the model.

        :param read_csv_kwargs: Keyword arguments to pass to pandas `read_csv`.
        """

        annot_mat = cls()

        # If the delimiter is not specified, assume whitespace by default:
        if 'sep' not in read_csv_kwargs and 'delimiter' not in read_csv_kwargs:
            read_csv_kwargs['delim_whitespace'] = True

        try:
            annot_df = pd.read_csv(annot_file, **read_csv_kwargs)
        except Exception as e:
            raise e

        annot_mat._snps = annot_df['SNP'].values
        annot_mat._chromosome = annot_df['CHR'][0]

        annot_df = annot_df.drop(['CHR', 'SNP', 'BP', 'CM', 'base'], axis=1)

        annot_mat._annotations = np.array(annot_df.columns)
        annot_mat._binary_annotations = np.array([c for c in annot_df.columns if len(annot_df[c].unique()) == 2])

        annot_mat._annot_matrix = annot_df.values

        return annot_mat

    @property
    def shape(self):
        return self.n_snps, self.n_annotations

    @property
    def n_snps(self):
        return len(self.snps)

    @property
    def chromosome(self):
        return self._chromosome

    @property
    def snps(self):
        return self._snps

    @property
    def n_annotations(self):
        return len(self.annotations)

    @property
    def binary_annotations(self):
        return self._binary_annotations

    @property
    def annotations(self):
        return self._annotations

    def annotation_index(self, annot):
        return list(self.annotations).index(annot)

    def values(self, add_intercept=False):
        """
        Returns the annotation matrix.
        :param add_intercept: Adds a base annotation corresponding to the intercept.
        """
        if add_intercept:
            return np.hstack([np.ones((self._annot_matrix.shape[0], 1)), self._annot_matrix])
        else:
            return self._annot_matrix

    def filter_snps(self, extract_snps=None, extract_file=None):
        """
        Filter variants from the annotation matrix. User must specify
        either a list of variants to extract or the path to a file
        with the list of variants to extract.

        :param extract_snps: A list (or array) of SNP IDs to keep in the annotation matrix.
        :param extract_file: The path to a file with the list of variants to extract.
        """

        assert extract_snps is not None or extract_file is not None

        if extract_file is not None:
            from .parsers.misc_parsers import read_snp_filter_file
            extract_snps = read_snp_filter_file(extract_file)

        arr_idx = intersect_arrays(self._snps, extract_snps, return_index=True)

        self._snps = self._snps[arr_idx]
        self._annot_matrix = self._annot_matrix[arr_idx, :]

    def filter_annotations(self, keep_annotations):
        """
        Filter the list of annotations in the matrix.
        :param keep_annotations: A list or vector of annotations to keep.
        """

        ann_idx = intersect_arrays(self._annotations, keep_annotations, return_index=True)

        self._annotations = self._annotations[ann_idx]
        self._annot_matrix = self._annot_matrix[:, ann_idx]

        self._binary_annotations = intersect_arrays(self._annotations, self._binary_annotations)

    def get_binary_annotation_index(self, bin_annot):
        """
        Get the indices of all SNPs that belong to binary annotation `bin_annot`
        """
        assert bin_annot in self._binary_annotations
        return np.where(self._annot_matrix[:, self.annotation_index(bin_annot)] == 1)[0]



import os.path as osp


def tgp_eur_data_path():
    """
    Return the path of the attached 1000G sample data for
    European individuals (N=378) and chromosome 22 (p=15938)
    """
    return osp.join(osp.dirname(osp.dirname(__file__)), 'data/1000G_eur_chr22')


def ukb_height_fastGWA_path():
    """
    Return the path of the attached GWAS summary statistics file
    for standing height. The file contains summary statistics for
    HapMap3 variants on CHR22 and is a snapshot of the summary statistics
    published on the fastGWA database:
    https://yanglab.westlake.edu.cn/data/fastgwa_data/UKB/50.v1.1.fastGWA.gz
    """
    return osp.join(osp.dirname(osp.dirname(__file__)), 'data/ukb_height_chr22.fastGWA.gz')



# cython: linetrace=False
# cython: profile=False
# cython: binding=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: language_level=3
# cython: infer_types=True

from libc.math cimport exp
import numpy as np
cimport numpy as np


def zarr_islice(arr, start=None, end=None):

    """
    This is copied from the official, but not yet released implementation of
    i_slice in Zarr codebase:
    https://github.com/zarr-developers/zarr-python/blob/e79e75ca8f07c95a5deede51f7074f699aa41149/zarr/core.py#L463
    :param arr: A Zarr array
    :param start: Start index
    :param end: End index
    """

    if len(arr.shape) == 0:
        # Same error as numpy
        raise TypeError("iteration over a 0-d array")
    if start is None:
        start = 0
    if end is None or end > arr.shape[0]:
        end = arr.shape[0]

    cdef unsigned int j, chunk_size = arr.chunks[0]
    chunk = None

    for j in range(start, end):
        if j % chunk_size == 0:
            chunk = arr[j: j + chunk_size]
        elif chunk is None:
            chunk_start = j - j % chunk_size
            chunk_end = chunk_start + chunk_size
            chunk = arr[chunk_start:chunk_end]
        yield chunk[j % chunk_size]


cpdef find_ld_block_boundaries(long[:] pos, long[:, :] block_boundaries):

    cdef unsigned int i, j, ldb_idx, block_start, block_end, B = len(block_boundaries), M = len(pos)
    cdef long[:] v_min = np.zeros_like(pos, dtype=np.int)
    cdef long[:] v_max = M*np.ones_like(pos, dtype=np.int)

    for i in range(M):

        # Find the positional boundaries for SNP i:
        for ldb_idx in range(B):
            if block_boundaries[ldb_idx, 0] <= pos[i] < block_boundaries[ldb_idx, 1]:
                block_start, block_end = block_boundaries[ldb_idx, 0], block_boundaries[ldb_idx, 1]
                break

        for j in range(i, M):
            if pos[j] >= block_end:
                v_max[i] = j
                break

        for j in range(i, -1, -1):
            if pos[j] < block_start:
                v_min[i] = j + 1
                break

    return np.array((v_min, v_max))


cpdef find_windowed_ld_boundaries(double[:] pos, double max_dist):

    cdef unsigned int i, j, M = len(pos)
    cdef long[:] v_min = np.zeros_like(pos, dtype=np.int)
    cdef long[:] v_max = M*np.ones_like(pos, dtype=np.int)

    for i in range(M):

        for j in range(i, M):
            if pos[j] - pos[i] > max_dist:
                v_max[i] = j
                break

        for j in range(i, -1, -1):
            if pos[i] - pos[j] > max_dist:
                v_min[i] = j + 1
                break

    return np.array((v_min, v_max))


cpdef find_shrinkage_ld_boundaries(double[:] cm_pos,
                                   double genmap_Ne,
                                   int genmap_sample_size,
                                   double cutoff):
    """
    Find the LD boundaries for the shrinkage estimator of Wen and Stephens (2010)
    
    :param cm_pos: A vector with the position of each genetic variant in centi Morgan.
    :param genmap_Ne: The effective population size for the genetic map sample.
    :param genmap_sample_size: The sample size used to estimate the genetic map.
    :param cutoff: The threshold below which we set the shrinkage factor to zero.
    """

    cdef unsigned int i, j, M = len(cm_pos)
    cdef long[:] v_min = np.zeros_like(cm_pos, dtype=np.int)
    cdef long[:] v_max = M*np.ones_like(cm_pos, dtype=np.int)

    # The multiplicative term for the shrinkage factor
    # The shrinkage factor is 4 * Ne * (rho_ij/100) / (2*m)
    # where Ne is the effective population size and m is the sample size
    # for the genetic map and rho_ij is the distance between SNPs i and j
    # in centi Morgan.
    # Therefore, the multiplicative term that we need to apply
    # to the distance between SNPs is: 4*Ne/(200*m), which is equivalent to 0.02*Ne/m
    # See also: https://github.com/stephenslab/rss/blob/master/misc/get_corr.R
    # and Wen and Stephens (2010)
    cdef double mult_term = 0.02 * genmap_Ne / genmap_sample_size

    for i in range(M):

        for j in range(i, M):
            if exp(-mult_term*(cm_pos[j] - cm_pos[i])) < cutoff:
                v_max[i] = j
                break

        for j in range(i, -1, -1):
            if exp(-mult_term*(cm_pos[i] - cm_pos[j])) < cutoff:
                v_min[i] = j + 1
                break

    return np.array((v_min, v_max))



# cython: linetrace=False
# cython: profile=False
# cython: binding=False
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: language_level=3
# cython: infer_types=True


import zarr
import os.path as osp
import numpy as np
cimport numpy as np
import pandas as pd
from magenpy.stats.ld.c_utils import zarr_islice


cdef class LDMatrix:

    cdef public:
        object _zarr
        bint in_memory
        list _data
        unsigned int index, _n_elements
        np.ndarray _ld_boundaries  # For caching
        np.ndarray _mask

    def __init__(self, zarr_arr):

        assert isinstance(zarr_arr, zarr.Array)

        self._zarr = zarr_arr

        self._data = None
        self.in_memory = False
        self.index = 0

        self._ld_boundaries = None
        self._mask = None
        self._n_elements = self.shape[0]

    @classmethod
    def from_path(cls, ld_store_path):
        """
        Initialize an `LDMatrix` object from a Zarr array store.
        :param ld_store_path: The path to the Zarr array store on the filesystem.
        """

        if '.zarray' in ld_store_path:
            ld_store_path = osp.dirname(ld_store_path)

        if osp.isfile(osp.join(ld_store_path, '.zarray')):
            ldm = zarr.open(ld_store_path)
            return cls(ldm)
        else:
            raise FileNotFoundError

    @classmethod
    def from_dir(cls, ld_store_path):
        """
        Initialize an `LDMatrix` object from a Zarr array store. See also `.from_path`
        :param ld_store_path: The path to the Zarr array store on the filesystem.
        """
        return cls.from_path(ld_store_path)

    @property
    def n_elements(self):
        """
        The number of non-masked elements in the LD matrix. For the full number
        see `.shape`.
        """
        return self._n_elements

    @property
    def n_snps(self):
        """
        The number of SNPs in the LD matrix. See also `.n_elements`
        """
        return self.n_elements

    @property
    def shape(self):
        """
        The shape (dimensions) of the LD matrix.
        """
        return self._zarr.shape

    @property
    def store(self):
        """
        The Zarr array store object.
        """
        return self._zarr.store

    @property
    def z_array(self):
        """
        The Zarr array
        """
        return self._zarr

    @property
    def chunks(self):
        """
        The chunks of the Zarr array
        """
        return self._zarr.chunks

    @property
    def chunk_size(self):
        """
        The chunk size of the Zarr array.
        """
        return self.chunks[0]

    @property
    def chromosome(self):
        """
        The chromosome for which this LD matrix was calculated.
        """
        return self.get_store_attr('Chromosome')

    @property
    def ld_estimator(self):
        """
        The LD estimator
        """
        return self.get_store_attr('LD estimator')

    @property
    def estimator_properties(self):
        """
        Properties of the LD estimator
        """
        return self.get_store_attr('Estimator properties')

    @property
    def snps(self):
        """
        The SNPs included in the LD matrix.
        """

        z_snps = np.array(self.get_store_attr('SNP'))

        if self._mask is not None:
            return z_snps[self._mask]
        else:
            return z_snps

    @property
    def ld_boundaries(self):
        """
        The LD boundaries associated with each SNP.
        """
        if self._ld_boundaries is None:
            self._ld_boundaries = np.array(self.get_store_attr('LD boundaries'))
        return np.array(self._ld_boundaries)

    @property
    def sample_size(self):
        """
        The sample size used to compute the LD matrix.
        """
        return self.get_store_attr('Sample size')

    @property
    def a1(self):
        """
        The alternative allele for which we count mutations at each SNP
        """

        a1 = np.array(self.get_store_attr('A1'))

        if self._mask is not None:
            return a1[self._mask]
        else:
            return a1

    @property
    def a2(self):
        """
        The reference allele
        """

        a2 = np.array(self.get_store_attr('A2'))

        if self._mask is not None:
            return a2[self._mask]
        else:
            return a2

    @property
    def maf(self):
        """
        The minor allele frequency (MAF) of each SNP in the LD matrix.
        """

        maf = np.array(self.get_store_attr('MAF'))

        if self._mask is not None:
            return maf[self._mask]
        else:
            return maf

    @property
    def bp_position(self):
        """
        The base pair position of each SNP in the LD matrix.
        """

        bp = np.array(self.get_store_attr('BP'))

        if self._mask is not None:
            return bp[self._mask]
        else:
            return bp

    @property
    def cm_position(self):
        """
        The centi Morgan position of each SNP in the LD matrix.
        """

        cm = np.array(self.get_store_attr('cM'))

        if cm is None:
            return None
        elif self._mask is not None:
            return cm[self._mask]
        else:
            return cm

    @property
    def ld_score(self):
        """
        The LD score of each SNP in the LD matrix.
        """

        ld_score = self.get_store_attr('LDScore')

        if ld_score is None:
            ld_score = self.compute_ld_scores()
            if self._mask is None:
                self.set_store_attr('LDScore', ld_score.tolist())
        else:
            ld_score = np.array(ld_score)

        if self._mask is not None:
            return ld_score[self._mask]
        else:
            return ld_score

    def filter_snps(self, extract_snps=None, extract_file=None):
        """
        Filter the LDMatrix to a subset of SNPs.
        :param extract_snps: A list or array of SNP IDs to keep.
        :param extract_file: A file containing the SNP IDs to keep.
        """

        assert extract_snps is not None or extract_file is not None

        if extract_snps is None:
            from .parsers.misc_parsers import read_snp_filter_file
            extract_snps = read_snp_filter_file(extract_file)

        from .utils.compute_utils import intersect_arrays

        extract_index = intersect_arrays(np.array(self.get_store_attr('SNP')),
                                         extract_snps,
                                         return_index=True)

        new_mask = np.zeros(self.shape[0], dtype=bool)
        new_mask[extract_index] = True

        self.set_mask(new_mask)

    def get_mask(self):
        """
        Get the mask used to hide/remove some SNPs from the LD matrix.
        """
        if self._mask is not None:
            return np.array(self._mask)

    def set_mask(self, mask):
        """
        Set the mask to hide/remove some SNPs from the LD matrix.
        :param mask: A boolean numpy array indicating whether to keep each SNP
        or not.
        """

        self._mask = mask

        if mask is None:
            # Update the number of elements:
            self._n_elements = self.shape[0]
        else:
            # Update the number of elements:
            self._n_elements = mask.sum()

            # Load the LD boundaries:
            ld_bounds = self.ld_boundaries

            # If the data is already in memory, reload:
            if self.in_memory:
                self.load(force_reload=True)

    def get_masked_boundaries(self):
        """
        Return the LD boundaries after applying the mask
        If the mask is not set, return the original boundaries
        """

        curr_ld_bounds = self.ld_boundaries

        if self._mask is None:
            return curr_ld_bounds
        else:
            # Number of excluded elements up to (and including) position i
            n_excluded = np.cumsum(~self._mask)
            # Number of excluded elements up to (not including) position i
            n_excluded_before = n_excluded - (~self._mask).astype(int)

            # New start position:
            start_pos = curr_ld_bounds[0, :] - n_excluded_before[curr_ld_bounds[0, :]]
            # New end position:
            end_pos = curr_ld_bounds[1, :] - n_excluded[curr_ld_bounds[1, :] - 1]

            # Return masked boundaries array:
            return np.array([start_pos[self._mask], end_pos[self._mask]])

    def to_snp_table(self, col_subset=None):
        """
        Return a table of the SNP attributes for SNPs in the LD matrix.
        """

        col_subset = col_subset or ['CHR', 'SNP', 'POS', 'A1', 'A2', 'MAF', 'LDScore']

        table = pd.DataFrame({'SNP': self.snps})

        for col in col_subset:
            if col == 'CHR':
                table['CHR'] = self.chromosome
            if col == 'POS':
                table['POS'] = self.bp_position
            if col == 'cM':
                table['cM'] = self.cm_position
            if col == 'A1':
                table['A1'] = self.a1
            if col == 'A2':
                table['A2'] = self.a2
            if col == 'MAF':
                table['MAF'] = self.maf
            if col == 'LDScore':
                table['LDScore'] = self.ld_score

        return table[list(col_subset)]

    def to_csr_matrix(self):
        """
        Convert the Zarr-formatted LD matrix into a sparse CSR matrix.
        """

        # Concatenate the data (entries of the LD matrix):
        data = np.concatenate([np.array(x) for x in self])

        # Stitch together the rows and columns for each data point:
        bounds = self.get_masked_boundaries()
        window_sizes = bounds[1, :] - bounds[0, :]

        rows = np.concatenate([np.repeat(i, ws) for i, ws in enumerate(window_sizes)])
        cols = np.concatenate([np.arange(bounds[0, i], bounds[1, i]) for i in range(bounds.shape[1])])

        from scipy.sparse import csr_matrix

        return csr_matrix((data, (rows, cols)), shape=(self.n_elements, self.n_elements))

    def compute_ld_scores(self, annotation_matrix=None, corrected=True):
        """
        Computes the LD scores for all SNPs in the LD matrix.
        :param annotation_matrix: A matrix of annotations for each variant for which to aggregate the LD scores.
        :param corrected: Use the sample-size corrected estimator for the squared Pearson correlation.
         See Bulik-Sullivan et al. 2015.
        """

        ld_scores = []
        cdef int n = self.sample_size

        for snp_ld, (start, end) in zip(self, self.get_masked_boundaries().T):

            ldsc = np.array(snp_ld) ** 2

            if corrected:
                ldsc = ldsc - (1. - ldsc) / (n - 2)

            if annotation_matrix is None:
                ld_scores.append(ldsc.sum())
            else:
                ld_scores.append((ldsc.reshape(-1, 1) * annotation_matrix[start: end, :]).sum(axis=0))

        return np.array(ld_scores)

    def store_size(self):
        """
        Returns the size of the compressed LD store in MB
        """
        return self.store.getsize() / 1024 ** 2

    def estimate_uncompressed_size(self):
        """
        Returns an estimate of size of the uncompressed LD matrix in MB
        If the array is masked, it returns a (rough) estimate of the size of the
        elements that will be loaded into memory.
        """
        ld_bounds = self.get_masked_boundaries()

        return (ld_bounds[1, :] - ld_bounds[0, :]).sum() * np.dtype(np.float64).itemsize / 1024 ** 2

    def get_store_attr(self, attr):
        """
        Get the attribute or metadata `attr` associated with the LD matrix.
        """
        try:
            return self._zarr.attrs[attr]
        except KeyError:
            print(f"Warning: Attribute '{attr}' is not set!")
            return None

    def set_store_attr(self, attr, value):
        """
        Set the attribute or metadata `attr` associated with the LD matrix.
        """
        try:
            self._zarr.attrs[attr] = value
        except Exception as e:
            raise e

    def load(self, start=0, end=None, force_reload=False):
        """
        Load the LD matrix from disk to memory.
        :param start: The start row position.
        :param end: The end row position.
        :param force_reload: If True, it will reload the data even if it was already loaded.
        """

        if self.in_memory and not force_reload:
            return

        if end is None:
            end = self.shape[0]
            if start == 0:
                self.in_memory = True

        cdef:
            unsigned int i
            long[:, ::1] ld_bounds = self.ld_boundaries

        self._data = []

        if self._mask is None:

            for d in zarr_islice(self._zarr, start, end):
                # TODO: Figure out a way to get around copying
                self._data.append(d.copy())

        else:

            for i, d in enumerate(zarr_islice(self._zarr, start, end), start):
                if self._mask[i]:
                    bound_start, bound_end = ld_bounds[:, i]
                    self._data.append(d[self._mask[bound_start: bound_end]].copy())
                else:
                    self._data.append(np.array([np.nan]))

    def release(self):
        """
        Release the LD data from memory.
        """
        self._data = None
        self.in_memory = False
        self.index = 0

    def iterate_chunks(self):
        """
        Iterate over chunks of the LD matrix.
        TODO: Incorporate the mask into the chunk iterator?
        """
        for i in range(self.shape[0] // self.chunk_size + 1):
            yield self.z_array[i*self.chunk_size:(i + 1)*self.chunk_size]

    def __getstate__(self):
        return self.store.path, self.in_memory, self._mask

    def __setstate__(self, state):

        path, in_mem, mask = state

        self._zarr = zarr.open(path)

        if mask is None:
            if in_mem:
                self.load()
        else:
            self.in_memory = in_mem
            self.set_mask(mask)

    def __len__(self):
        return self._n_elements

    def __getitem__(self, item):
        if self.in_memory:
            return self._data[item]
        else:
            return self._zarr[item]

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):

        cdef int i, curr_chunk, index_chunk

        if self.index == 0:
            curr_chunk = -1
        else:
            curr_chunk = (self.index - 1) // self.chunk_size

        if self._mask is not None:

            try:
                if not self._mask[self.index]:
                    for i in range(0, self.shape[0] - self.index + 1):
                        if self._mask[self.index + i]:
                            break

                    self.index += i
            except IndexError:
                # Reached the end of the array
                self.index = self.shape[0]
                pass

        if self.index == self.shape[0]:
            self.index = 0
            raise StopIteration

        cdef double[::1] next_item

        if self.in_memory:
            next_item = self._data[self.index]
        else:
            index_chunk = self.index // self.chunk_size
            if index_chunk > curr_chunk:
                self.load(start=index_chunk * self.chunk_size, end=(index_chunk + 1) * self.chunk_size)

            next_item = self._data[self.index % self.chunk_size]

        self.index += 1

        return next_item



import numpy as np
import pandas as pd
import collections
import six


def generate_slice_dictionary(vec):
    """
    This utility function takes a sorted vector (e.g. numpy array),
    identifies the unique elements and generates a dictionary of slices
    delineating the start and end positions of each element in the vector.

    :param vec: A numpy array
    """

    vals, idx = np.unique(vec, return_index=True)
    idx_sort = np.argsort(idx)

    vals = vals[idx_sort]
    idx = idx[idx_sort]

    d = {}

    for i in range(len(idx)):
        try:
            d[vals[i]] = slice(idx[i], idx[i + 1])
        except IndexError:
            d[vals[i]] = slice(idx[i], len(vec))

    return d


def intersect_arrays(arr1, arr2, return_index=False):
    """
    This utility function takes two arrays and returns the shared
    elements (intersection) between them. If return_index is set to True,
    it returns the index of shared elements in the first array.

    :param arr1: The first array
    :param arr2: The second array
    :param return_index: Return the index of shared elements in the first array
    :return:
    """

    # NOTE: For best and consistent results, we cast all data types to `str`
    # for now. May need a smarter solution in the future.
    common_elements = pd.DataFrame({'ID': arr1}, dtype=str).reset_index().merge(
        pd.DataFrame({'ID': arr2}, dtype=str)
    )

    if return_index:
        return common_elements['index'].values
    else:
        return common_elements['ID'].values


def iterable(arg):
    return (
        isinstance(arg, collections.Iterable)
        and not isinstance(arg, six.string_types)
    )



import errno
import os
import os.path as osp
import subprocess
import glob
import psutil


def available_cpu():
    return psutil.cpu_count() - 1


def valid_url(path):

    import requests

    try:
        r = requests.head(path)
        return r.status_code == requests.codes.ok
    except Exception:
        return False


def is_cmd_tool(name):
    """
    Check whether `name` is on PATH and marked as executable.
    From: https://stackoverflow.com/a/34177358
    """
    from shutil import which

    return which(name) is not None


def makedir(dirs):
    """
    Create directories on the filesystem, recursively.
    """

    if isinstance(dirs, str):
        dirs = [dirs]

    for dir in dirs:
        try:
            os.makedirs(dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise


def get_filenames(path, extension=None):
    """
    Obtain valid and full path names given the provided `path` or prefix and extensions.

    :param path: A string with the path prefix or full path.
    :param extension: The extension for the class of files to search for.
    """

    if osp.isdir(path):
        if extension:
            if osp.isfile(osp.join(path, extension)):
                return [path]
            else:
                return [f for f in glob.glob(osp.join(path, '*/'))
                        if extension in f or osp.isfile(osp.join(f, extension))]
        else:
            return glob.glob(osp.join(path, '*'))
    else:
        if extension is None:
            return glob.glob(path + '*')
        elif extension in path:
            return glob.glob(path)
        elif osp.isfile(path + extension):
            return [path + extension]
        else:
            return (
                    glob.glob(osp.join(path, '*' + extension + '*')) +
                    glob.glob(osp.join(path, extension + '*')) +
                    glob.glob(path + '*' + extension + '*')
            )


def run_shell_script(cmd):
    """
    Run the shell script given the command prompt in `cmd`.
    :param cmd: A string with the shell command to run.
    """

    result = subprocess.run(cmd, shell=True, capture_output=True)

    if result.stderr:
        raise subprocess.CalledProcessError(
            returncode=result.returncode,
            cmd=result.args,
            stderr=result.stderr
        )

    return result


def delete_temp_files(prefix):
    """
    Delete temporary files with the given `prefix`.
    """
    for f in glob.glob(f"{prefix}*"):
        try:
            os.remove(f)
        except Exception as e:
            continue




import pandas as pd


class SumstatsParser(object):
    """
    A genetic summary statistics parser class.
    """

    def __init__(self, col_name_converter=None):
        """
        :param col_name_converter: A dictionary mapping column names
        in the original table to magenpy's column names for the various
        summary statistics.
        """
        self.col_name_converter = col_name_converter

    def parse(self, file_name, drop_na=True, **read_csv_kwargs):
        """
        Parse a summary statistics file.
        :param file_name: The path to the summary statistics file.
        :param drop_na: Drop any entries with missing values.
        """

        # If the delimiter is not specified, assume whitespace by default:
        if 'sep' not in read_csv_kwargs and 'delimiter' not in read_csv_kwargs:
            read_csv_kwargs['delim_whitespace'] = True

        df = pd.read_csv(file_name, **read_csv_kwargs)

        if drop_na:
            df = df.dropna()

        if self.col_name_converter is not None:
            df.rename(columns=self.col_name_converter, inplace=True)

        return df


class plinkSumstatsParser(SumstatsParser):
    """
    A parser for plink GWAS summary statistics files.
    """

    def __init__(self, col_name_converter=None):
        """
        :param col_name_converter: A dictionary mapping column names
        in the original table to magenpy's column names for the various
        summary statistics.
        """

        super().__init__(col_name_converter)

        self.col_name_converter = self.col_name_converter or {}

        self.col_name_converter.update(
            {
                '#CHROM': 'CHR',
                'ID': 'SNP',
                'P': 'PVAL',
                'OBS_CT': 'N',
                'A1_FREQ': 'MAF',
                'T_STAT': 'Z',
                'Z_STAT': 'Z'
            }
        )

    def parse(self, file_name, drop_na=True, **read_csv_kwargs):
        """
        Parse a summary statistics file.
        :param file_name: The path to the summary statistics file.
        :param drop_na: Drop any entries with missing values.
        """

        df = super().parse(file_name, drop_na=drop_na, **read_csv_kwargs)

        try:
            df['A2'] = df.apply(lambda x: [x['ALT1'], x['REF']][x['A1'] == x['ALT1']], axis=1)
        except KeyError:
            print("Warning: the reference allele A2 could not be inferred "
                  "from the summary statistics file!")

        return df


class COJOSumstatsParser(SumstatsParser):
    """
    A parser for COJO GWAS summary statistics files.
    """

    def __init__(self, col_name_converter=None):
        """
        :param col_name_converter: A dictionary mapping column names
        in the original table to magenpy's column names for the various
        summary statistics.
        """
        super().__init__(col_name_converter)

        self.col_name_converter = self.col_name_converter or {}

        self.col_name_converter.update(
            {
                'freq': 'MAF',
                'b': 'BETA',
                'se': 'SE',
                'p': 'PVAL'
            }
        )


class fastGWASumstatsParser(SumstatsParser):
    """
    A parser for fastGWA summary statistics files
    """

    def __init__(self, col_name_converter=None):
        """
        :param col_name_converter: A dictionary mapping column names
        in the original table to magenpy's column names for the various
        summary statistics.
        """
        super().__init__(col_name_converter)

        self.col_name_converter = self.col_name_converter or {}

        self.col_name_converter.update(
            {
                'AF1': 'MAF',
                'P': 'PVAL'
            }
        )



import numpy as np
import pandas as pd


def read_snp_filter_file(filename, snp_id_col=0):
    """
    Read plink-style file listing variant IDs.
    The file should not have a header and only has a single column.
    """

    try:
        keep_list = pd.read_csv(filename, sep="\t", header=None).values[:, snp_id_col]
    except Exception as e:
        raise e

    return keep_list


def read_sample_filter_file(filename):
    """
    Read plink-style file listing sample IDs.
    The file should not have a header, be tab-separated, and has two
    columns corresponding to Family ID (FID) and Individual ID (IID).
    You may also pass a file with a single-column of Individual IDs instead.
    """

    keep_list = pd.read_csv(filename, sep="\t", header=None).values

    if keep_list.shape[1] == 1:
        return keep_list[:, 0]
    elif keep_list.shape[1] == 2:
        return keep_list[:, 1]


def parse_ld_block_data(ldb_file_path):
    """
    This function takes a path to a file with the LD blocks
    and returns a dictionary with the chromosome ID and a list of the
    start and end positions for the blocks in that chromosome.
    The parser assumes that the LD block files have the ldetect format:
    https://bitbucket.org/nygcresearch/ldetect-data/src/master/

    :param ldb_file_path: The path to the LD blocks file
    """

    ld_blocks = {}

    df = pd.read_csv(ldb_file_path, delim_whitespace=True,
                     dtype={'chr': str, 'start': np.int64, 'end': np.int64})

    if df.isnull().values.any():
        raise ValueError("The LD block data contains missing information. This may result in invalid "
                         "LD boundaries. Please fix the LD block files before continuing!")

    for chrom in df['chr'].unique():
        ld_blocks[int(chrom.replace('chr', ''))] = df.loc[df['chr'] == chrom, ['start', 'stop']].values

    return ld_blocks


def parse_cluster_assignment_file(cluster_assignment_file):
    """
    Parses a file that maps each individual in the sample table to a cluster,
    and returns the pandas dataframe. The expected file should be whitespace delimited
    and contain three columns: FID, IID, and Cluster

    :param cluster_assignment_file: The path to the cluster assignment file.
    """
    try:
        clusters = pd.read_csv(cluster_assignment_file, delim_whitespace=True)
        clusters.columns = ['FID', 'IID', 'Cluster']
    except Exception as e:
        raise e

    return clusters




import warnings
import pandas as pd
import numpy as np
from magenpy.utils.compute_utils import intersect_arrays


class SumstatsTable(object):

    def __init__(self, ss_table: pd.DataFrame):
        self.table: pd.DataFrame = ss_table

        assert all([col in self.table.columns for col in ('SNP', 'A1')])

    @property
    def shape(self):
        return self.table.shape

    @property
    def chromosome(self):
        chrom = self.chromosomes
        if chrom is not None:
            if len(chrom) == 1:
                return chrom[0]

    @property
    def chromosomes(self):
        if 'CHR' in self.table.columns:
            return self.table['CHR'].unique()

    @property
    def m(self):
        return self.n_snps

    @property
    def n_snps(self):
        return len(self.table)

    @property
    def snps(self):
        return self.table['SNP'].values

    @property
    def a1(self):
        return self.table['A1'].values

    @property
    def a2(self):
        return self.get_col('A2')

    @property
    def ref_allele(self):
        return self.a2

    @property
    def alt_allele(self):
        return self.a1

    @property
    def effect_allele(self):
        return self.a1

    @property
    def bp_pos(self):
        return self.get_col('POS')

    @property
    def maf(self):
        return self.get_col('MAF')

    @property
    def maf_var(self):
        return 2.*self.maf*(1. - self.maf)

    @property
    def n(self):
        return self.get_col('N')

    @property
    def n_per_snp(self):
        return self.get_col('N')

    @property
    def beta_hat(self):
        return self.get_col('BETA')

    @property
    def marginal_beta(self):
        return self.beta_hat

    @property
    def z_score(self):
        z = self.get_col('Z')
        if z is not None:
            return z
        else:
            beta = self.beta_hat
            se = self.se

            if beta is not None and se is not None:
                self.table['Z'] = beta / se
                return self.table['Z'].values
            else:
                raise Exception("Z-score statistic is not available!")

    @property
    def standard_error(self):
        return self.get_col('SE')

    @property
    def se(self):
        return self.standard_error

    @property
    def pval(self):
        p = self.get_col('PVAL')

        if p is not None:
            return p
        else:
            from scipy import stats
            self.table['PVAL'] = 2.*stats.norm.sf(np.abs(self.z_score))
            return self.table['PVAL'].values

    @property
    def p_value(self):
        return self.pval

    @property
    def log10_p_value(self):
        """
        Computes -log10(p_value).
        May be useful for Manhattan plots.
        """
        return -np.log10(self.pval)

    @property
    def effect_sign(self):
        """
        Return the sign for the effect size (1 for positive effect, -1 for negative effect)
        of each genetic marker.
        """

        signed_statistic = self.beta_hat or self.z_score

        if signed_statistic is not None:
            return np.sign(signed_statistic)
        else:
            raise Exception("No signed statistic to extract the sign from!")

    def match(self, reference_table, correct_flips=True):
        """
        Match the summary statistics table with a reference table,
        correcting for potential flips in the effect allele.

        :param reference_table: The SNP table to use as a reference. Must be a pandas
        table with at least three columns: SNP, A1, A2.
        :param correct_flips: If True, correct the direction of effect size
         estimates if the effect allele is reversed.
        """

        from magenpy.utils.model_utils import merge_snp_tables

        self.table = merge_snp_tables(ref_table=reference_table[['SNP', 'A1', 'A2']],
                                      alt_table=self.table,
                                      how='inner',
                                      correct_flips=correct_flips)

    def filter_by_allele_frequency(self, min_maf=None, min_mac=None):
        """
        Filter variants by minimum minor allele frequency or allele count
        :param min_maf: Minimum minor allele frequency
        :param min_mac: Minimum minor allele count
        """

        if min_mac or min_maf:
            maf = self.maf
            n = self.n_per_snp
        else:
            return

        keep_flag = None

        if min_mac and n and maf:
            mac = (2*maf*n).astype(np.int64)
            keep_flag = (mac >= min_mac) & ((2*n - mac) >= min_mac)

        if min_maf and maf:

            maf_cond = (maf >= min_maf) & (1. - maf >= min_maf)
            if keep_flag is not None:
                keep_flag = keep_flag & maf_cond
            else:
                keep_flag = maf_cond

        if keep_flag is not None:
            self.filter_snps(extract_index=np.where(keep_flag)[0])

    def filter_snps(self, extract_snps=None, extract_file=None, extract_index=None):
        """
        Filter the summary statistics table to a subset of SNPs.
        :param extract_snps: A list or array of SNP IDs to keep.
        :param extract_file: A file containing the SNP IDs to keep.
        :param extract_index: A list or array of the indices of SNPs to retain.
        """

        assert extract_snps is not None or extract_file is not None or extract_index is not None

        if extract_file:
            from .parsers.misc_parsers import read_snp_filter_file
            extract_snps = read_snp_filter_file(extract_file)

        if extract_snps is not None:
            extract_index = intersect_arrays(self.snps, extract_snps, return_index=True)

        if extract_index is not None:
            self.table = self.table.iloc[extract_index, ].reset_index(drop=True)
        else:
            raise Exception("To filter a summary statistics table, you must provide "
                            "the list of SNPs, a file containing the list of SNPs, or a list of indices to retain.")

    def drop_duplicates(self):
        """
        Drop duplicated SNP rsIDs.
        """

        self.table = self.table.drop_duplicates(subset='SNP', keep=False)

    def get_col(self, col_name):
        """
        Returns a particular summary statistic or column from the summary statistics 
        table.
        :param col_name: The name of the column
        """
        if col_name in self.table.columns:
            return self.table[col_name].values

    def get_chisq_statistic(self):
        """
        Obtain the Chi-Squared statistic
        """
        chisq = self.get_col('CHISQ')

        if chisq is not None:
            return chisq
        else:
            z = self.z_score
            if z is not None:
                self.table['CHISQ'] = z**2
            else:
                p_val = self.p_value
                if p_val is not None:
                    from scipy.stats import chi2

                    self.table['CHISQ'] = chi2.ppf(1. - p_val, 1)
                else:
                    raise ValueError("Chi-Squared statistic is not available!")

        return self.table['CHISQ'].values

    def get_snp_pseudo_corr(self):
        """
        Computes the pseudo-correlation coefficient (standardized beta) between the SNP and
        the phenotype (X_jTy / N) from GWAS summary statistics.
        Uses Equation 15 in Mak et al. 2017
        beta =  z_j / sqrt(n - 1 + z_j^2)
        Where z_j is the marginal GWAS Z-score
        """

        zsc = self.z_score
        n = self.n

        if zsc is not None:
            if n is not None:
                return zsc / (np.sqrt(n - 1 + zsc**2))
            else:
                raise Exception("Sample size is not available!")
        else:
            raise Exception("Z-scores are not available!")
        
    def get_yy_per_snp(self):
        """
        Computes the quantity (y'y)_j/n_j following SBayesR (Lloyd-Jones 2019) and Yang et al. (2012).
        (y'y)_j/n_j is the empirical variance for continuous phenotypes and may be estimated
        from GWAS summary statistics by re-arranging the equation for the
        squared standard error:

        SE(b_j)^2 = (Var(y) - Var(x_j)*b_j^2) / (Var(x)*n)

        Which gives the following estimate:

        (y'y)_j / n_j = (n_j - 2)*SE(b_j)^2 + b_j^2

        """
        
        b = self.beta_hat
        se = self.standard_error
        n = self.n

        if n is not None:
            if b is not None:
                if se is not None:
                    return (n - 2)*se**2 + b**2
                else:
                    raise Exception("Standard errors are not available!")
            else:
                raise Exception("Marginal betas are not available!")
        else:
            raise Exception("Sample size per SNP is not available!")

    def split_by_chromosome(self, snps_per_chrom=None):
        """
        Split the summary statistics table by chromosome, so that we would
        have a separate `SumstatsTable` table for each chromosome.
        :param snps_per_chrom: A dictionary where the keys are the chromosome number 
        and the value is an array or list of SNPs on that chromosome.
        """

        if 'CHR' in self.table.columns:
            chrom_tables = self.table.groupby('CHR')
            return {
                c: SumstatsTable(chrom_tables.get_group(c))
                for c in chrom_tables.groups
            }
        elif snps_per_chrom is not None:
            chrom_dict = {
                c: SumstatsTable(pd.DataFrame({'SNP': snps}).merge(self.table))
                for c, snps in snps_per_chrom.items()
            }

            for c, ss_tab in chrom_dict.items():
                ss_tab.table['CHR'] = c

            return chrom_dict
        else:
            raise Exception("To split the summary statistics table by chromosome, "
                            "you must provide the a dictionary mapping chromosome number "
                            "to an array of SNPs `snps_per_chrom`.")

    def get_table(self, col_subset=None):
        """
        Get the summary statistics table or a subset of it.
        :param col_subset: A list corresponding to a subset of columns to return.
        """

        col_subset = col_subset or ['CHR', 'SNP', 'POS', 'A1', 'A2', 'MAF',
                                    'N', 'BETA', 'Z', 'SE', 'PVAL']

        # Because some of the quantities that the user needs may be need to be
        # computed, we separate the column subset into those that are already
        # present in the table and those that are not (but can still be computed
        # from other summary statistics):

        present_cols = list(set(col_subset).intersection(set(self.table.columns)))
        non_present_cols = list(set(col_subset) - set(present_cols))

        if len(present_cols) > 0:
            table = self.table[present_cols].copy()
        else:
            table = pd.DataFrame({c: [] for c in non_present_cols})

        for col in non_present_cols:

            if col == 'Z':
                table['Z'] = self.z_score
            elif col == 'PVAL':
                table['PVAL'] = self.p_value
            elif col == 'LOG10_PVAL':
                table['LOG10_PVAL'] = self.log10_p_value
            elif col == 'CHISQ':
                table['CHISQ'] = self.get_chisq_statistic()
            elif col == 'MAF_VAR':
                table['MAF_VAR'] = self.maf_var
            elif col == 'STD_BETA':
                table['STD_BETA'] = self.get_snp_pseudo_corr()
            else:
                warnings.warn(f"Column '{col}' is not available in the summary statistics table!")

        return table[list(col_subset)]

    def to_file(self, output_file, col_subset=None, **to_csv_kwargs):
        """
        Write the summary statistics table to file.

        TODO: Add a format argument to this method and allow the user to output summary statistics
        according to supported formats (e.g. COJO, plink, fastGWA, etc.).

        :param output_file: The path to the file where to write the summary statistics.
        :param col_subset: A subset of the columns to write to file.
        """

        if 'sep' not in to_csv_kwargs and 'delimiter' not in to_csv_kwargs:
            to_csv_kwargs['sep'] = '\t'

        if 'index' not in to_csv_kwargs:
            to_csv_kwargs['index'] = False

        table = self.get_table(col_subset)
        table.to_csv(output_file, **to_csv_kwargs)

    @classmethod
    def from_file(cls, sumstats_file, sumstats_format=None, parser=None, **parse_kwargs):
        """
        Initialize a summary statistics table from file. The user must provide either
        the format for the summary statistics file or the parser object
        (see parsers.sumstats_parsers).
        :param sumstats_file: The path to the summary statistics file.
        :param sumstats_format: The format for the summary statistics file. Currently,
        we explicitly support the following three formats formats: magenpy, plink, COJO.
        :param parser: An instance of SumstatsParser parser, implements basic parsing/conversion
        functionalities.
        :param parse_kwargs: arguments for the pandas `read_csv` function, such as the delimiter.
        """
        assert sumstats_format is not None or parser is not None

        from .parsers.sumstats_parsers import (
            SumstatsParser, plinkSumstatsParser, COJOSumstatsParser, fastGWASumstatsParser
        )

        if parser is None:
            if sumstats_format == 'magenpy':
                parser = SumstatsParser()
            elif sumstats_format == 'plink':
                parser = plinkSumstatsParser()
            elif sumstats_format == 'COJO':
                parser = COJOSumstatsParser()
            elif sumstats_format == 'fastGWA':
                parser = fastGWASumstatsParser()
            else:
                raise KeyError(f"Parsers for summary statistics format {sumstats_format} are not implemented!")

        sumstats_table = parser.parse(sumstats_file, **parse_kwargs)
        return cls(sumstats_table)



from setuptools import setup, Extension, find_packages
import numpy as np
import os

try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = None

# ------------------------------------------------------
# Cython dependencies:


# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


extensions = [
    Extension("magenpy.stats.ld.c_utils",
              sources=["magenpy/stats/ld/c_utils.pyx"],
              extra_compile_args=["-ffast-math"],
              include_dirs=[np.get_include()],
              ),
    Extension("magenpy.LDMatrix",
              sources=["magenpy/LDMatrix.pyx"],
              extra_compile_args=["-ffast-math"],
              include_dirs=[np.get_include()],
              )
]
if cythonize is not None:
    compiler_directives = {
        "language_level": 3,
        "embedsignature": True,
        'boundscheck': False,
        'wraparound': False,
        'nonecheck': False,
        'cdivision': True
    }
    extensions = cythonize(extensions, compiler_directives=compiler_directives)
else:
    extensions = no_cythonize(extensions)

# ------------------------------------------------------
# Read description/dependencies from file:

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")

with open("requirements-optional.txt") as fp:
    opt_requires = fp.read().strip().split("\n")

# ------------------------------------------------------

setup(
    name="magenpy",
    version="0.0.10",
    author="Shadi Zabad",
    author_email="shadi.zabad@mail.mcgill.ca",
    description="Modeling and Analysis of Statistical Genetics data in python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/shz9/magenpy",
    classifiers=[
        'Programming Language :: Python',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    package_dir={'': '.'},
    packages=find_packages(),
    package_data={'magenpy': ['data/*.bed', 'data/*.bim', 'data/*.fam',
                              'data/ukb_height_chr22.fastGWA.gz',
                              'config/*.ini']},
    scripts=['bin/magenpy_ld', 'bin/magenpy_simulate'],
    install_requires=install_requires,
    extras_require={'full': opt_requires},
    ext_modules=extensions,
    zip_safe=False
)



import pandas as pd


def parse_bim_file(plink_bfile):
    """
    From the plink documentation:
    https://www.cog-genomics.org/plink/1.9/formats#bim

        A text file with no header line, and one line per variant with the following six fields:

        - Chromosome code (either an integer, or 'X'/'Y'/'XY'/'MT'; '0' indicates unknown) or name
        - Variant identifier
        - Position in morgans or centimorgans (safe to use dummy value of '0')
        - Base-pair coordinate (1-based; limited to 231-2)
        - Allele 1 (corresponding to clear bits in .bed; usually minor)
        - Allele 2 (corresponding to set bits in .bed; usually major)

    :param plink_bfile:
    """

    if '.bim' not in plink_bfile:
        if '.bed' in plink_bfile:
            plink_bfile = plink_bfile.replace('.bed', '.bim')
        else:
            plink_bfile = plink_bfile + '.bim'

    bim_df = pd.read_csv(plink_bfile, delim_whitespace=True,
                         names=['CHR', 'SNP', 'cM', 'POS', 'A1', 'A2'],
                         dtype={
                             'CHR': int,
                             'SNP': str,
                             'cM': float,
                             'POS': int,
                             'A1': str,
                             'A2': str
                         })

    return bim_df


def parse_fam_file(plink_bfile):
    """
    From the plink documentation:
    https://www.cog-genomics.org/plink/1.9/formats#fam

        A text file with no header line, and one line per sample with the following six fields:

        - Family ID ('FID')
        - Within-family ID ('IID'; cannot be '0')
        - Within-family ID of father ('0' if father isn't in dataset)
        - Within-family ID of mother ('0' if mother isn't in dataset)
        - Sex code ('1' = male, '2' = female, '0' = unknown)
        - Phenotype value ('1' = control, '2' = case, '-9'/'0'/non-numeric = missing data if case/control)

    :param plink_bfile:
    """

    if '.fam' not in plink_bfile:
        if '.bed' in plink_bfile:
            plink_bfile = plink_bfile.replace('.bed', '.fam')
        else:
            plink_bfile = plink_bfile + '.fam'

    fam_df = pd.read_csv(plink_bfile,
                         delim_whitespace=True,
                         usecols=range(6),
                         names=['FID', 'IID', 'fatherID', 'motherID', 'sex', 'phenotype'],
                         dtype={'FID': str,
                                'IID': str,
                                'fatherID': str,
                                'motherID': str,
                                'sex': float,
                                'phenotype': float
                                },
                         na_values={
                             'phenotype': [-9.],
                             'sex': [0]
                         })

    if fam_df['phenotype'].isnull().all():
        fam_df.drop('phenotype', axis=1, inplace=True)

    if fam_df['sex'].isnull().all():
        fam_df.drop('sex', axis=1, inplace=True)

    return fam_df
